import logging
import os
import shutil
import subprocess
import tempfile

from ansible_galaxy import exceptions

log = logging.getLogger(__name__)


# TODO: split hg and git impls into sep methods, kind of a strategy pattern
def scm_archive_content(src, name, scm='git', version='HEAD'):
    """
    Archive a Galaxy Repository SCM repo locally

    Implementation originally adopted from the Ansible RoleRequirement
    """
    if scm not in ['hg', 'git']:
        raise exceptions.GalaxyClientError("- scm %s is not currently supported" % scm)

    tempdir = tempfile.mkdtemp()
    clone_cmd = [scm, 'clone', src, name]

    log.debug('clone_cmd: %s', clone_cmd)
    log.debug('clone tempdir: %s', tempdir)

    with open('/dev/null', 'w') as devnull:
        try:
            popen = subprocess.Popen(clone_cmd, cwd=tempdir, stdout=devnull, stderr=devnull)
        except Exception as exc:
            log.debug(exc, exc_info=True)
            raise exceptions.GalaxyClientError('error executing: "%s": %s' % (" ".join(clone_cmd), exc))
        rc = popen.wait()
    if rc != 0:
        raise exceptions.GalaxyClientError("- command %s failed in directory %s (rc=%s)" % (' '.join(clone_cmd), tempdir, rc))

    if scm == 'git' and version:
        checkout_cmd = [scm, 'checkout', version]
        with open('/dev/null', 'w') as devnull:
            try:
                popen = subprocess.Popen(checkout_cmd, cwd=os.path.join(tempdir, name), stdout=devnull, stderr=devnull)
            except (IOError, OSError) as exc:
                log.debug(exc, exc_info=True)
                raise exceptions.GalaxyClientError("error executing: %s" % " ".join(checkout_cmd))
            rc = popen.wait()
        if rc != 0:
            raise exceptions.GalaxyClientError("- command %s failed in directory %s (rc=%s)" % (' '.join(checkout_cmd), tempdir, rc))

    temp_file = tempfile.NamedTemporaryFile(delete=False,
                                            prefix='tmp-ansible-galaxy-content-archive-',
                                            suffix='.tar')

    log.debug('Archiving scm src %s to tar file: %s', src, temp_file.name)

    if scm == 'hg':
        archive_cmd = ['hg', 'archive', '--prefix', "%s/" % name]
        if version:
            archive_cmd.extend(['-r', version])
        archive_cmd.append(temp_file.name)
    if scm == 'git':
        archive_cmd = ['git', 'archive', '--prefix=%s/' % name, '--output=%s' % temp_file.name]
        if version:
            archive_cmd.append(version)
        else:
            archive_cmd.append('HEAD')

    with open('/dev/null', 'w') as devnull:
        popen = subprocess.Popen(archive_cmd, cwd=os.path.join(tempdir, name),
                                 stderr=devnull, stdout=devnull)
        rc = popen.wait()
    if rc != 0:
        raise exceptions.GalaxyClientError("- command %s failed in directory %s (rc=%s)" % (' '.join(archive_cmd), tempdir, rc))

    shutil.rmtree(tempdir, ignore_errors=True)
    return temp_file.name



import logging

import pytest

from ansible_galaxy import exceptions
from ansible_galaxy.fetch import galaxy_url
from ansible_galaxy.models.requirement_spec import RequirementSpec

log = logging.getLogger(__name__)

EXAMPLE_REPO_VERSION_LATEST = \
        {
            "id": 55616,
            "version": "1.1.0",
            "tag": "1.1.0",
            "commit_date": "2015-11-24T13:00:32-05:00",
            "commit_sha": None,
            "download_url": "https://github.com/atestuseraccount/ansible-role-logstash/archive/1.1.0.tar.gz",
            "url": "",
            "related": {},
            "summary_fields": {},
            "created": "2018-05-18T13:13:40.208012Z",
            "modified": "2018-05-18T13:13:40.208037Z",
            "active": None
        }

EXAMPLE_REPO_VERSIONS_LIST = \
    [
        {
            "id": 55617,
            "version": "1.0.6",
            "tag": "1.0.6",
            "commit_date": "2015-11-16T11:08:07-05:00",
            "commit_sha": None,
            "download_url": "https://github.com/atestuseraccount/ansible-role-logstash/archive/1.0.6.tar.gz",
            "url": "",
            "related": {},
            "summary_fields": {},
            "created": "2018-05-18T13:13:40.344064Z",
            "modified": "2018-05-18T13:13:40.344088Z",
            "active": None
        },
        EXAMPLE_REPO_VERSION_LATEST,
    ]


@pytest.fixture
def galaxy_url_fetch(galaxy_context):
    req_spec = RequirementSpec(namespace='some_namespace',
                               name='some_name',
                               version_spec='==9.3.245')

    galaxy_url_fetch = galaxy_url.GalaxyUrlFetch(requirement_spec=req_spec, galaxy_context=galaxy_context)
    log.debug('galaxy_url_fetch: %s', galaxy_url_fetch)

    return galaxy_url_fetch


def test_galaxy_url_fetch_find(galaxy_url_fetch, mocker):
    MockedGalaxyAPI = mocker.patch('ansible_galaxy.fetch.galaxy_url.GalaxyAPI', autospec=True)

    external_url = 'http://example.invalid/invalid/whatever'

    instance = MockedGalaxyAPI.return_value
    instance.lookup_repo_by_name.return_value = {'related': {'versions': 'http://example.invalid/foo'},
                                                 'external_url': external_url}
    instance.fetch_content_related.return_value = [{'version': '1.2.3'}, {'version': '9.3.245'}]

    res = galaxy_url_fetch.find()

    log.debug('res:%s', res)

    assert res['content']['galaxy_namespace'] == 'some_namespace'
    assert res['content']['repo_name'] == 'some_name'

    assert res['custom']['external_url'] == external_url


def test_galaxy_url_fetch_find_no_repo_data(galaxy_url_fetch, mocker):
    MockedGalaxyAPI = mocker.patch('ansible_galaxy.fetch.galaxy_url.GalaxyAPI', autospec=True)

    instance = MockedGalaxyAPI.return_value
    instance.lookup_repo_by_name.return_value = {}

    faux_server_url = 'http://galaxy.invalid/'
    instance.api_server = faux_server_url

    with pytest.raises(exceptions.GalaxyClientError, match='- sorry, some_namespace.some_name was not found on %s' % faux_server_url) as exc_info:
        galaxy_url_fetch.find()

    log.debug('exc_info:%s', exc_info)


def test_galaxy_url_fetch_fetch(galaxy_url_fetch, mocker):
    collection_path = '/dev/null/path/to/collection.tar.gz'

    mocked_download_fetch_url = mocker.patch('ansible_galaxy.fetch.galaxy_url.download.fetch_url', autospec=True)
    mocked_download_fetch_url.return_value = collection_path

    find_results = {'content': {'galaxy_namespace': 'some_namespace',
                                'repo_name': 'some_name'},
                    'custom': {'repo_data': {},
                               'external_url': 'http://example.invalid/invalid/whatever',
                               'repoversion': {'version': '9.3.245'}
                               },
                    }

    res = galaxy_url_fetch.fetch(find_results)

    log.debug('res:%s', res)

    assert isinstance(res, dict)
    assert res['archive_path'] == collection_path
    assert res['download_url'] == 'http://example.invalid/invalid/whatever/archive/9.3.245.tar.gz'
    assert res['fetch_method'] == 'galaxy_url'
    assert isinstance(res['content'], dict)
    assert res['content']['fetched_version'] == '9.3.245'


# Note that select_collection_version just gets the full version object
# from repoversions. It does not sort or compare versions aside from equality
@pytest.mark.parametrize("repoversions,version,expected", [
    ([], None, {}),
    ([{'version': '1.2.3'}],
     '1.2.3',
     {'version': '1.2.3'}),
    # the version requested isnt in repoversions, exp a {} result
    ([{'version': '1.2.4'}],
     '1.2.3',
     {}),
    # repoversions has 1.2.3 so expect the full repoversion object returned
    ([{'version': '1.2.2'},
      {'version': '1.2.3'},
      {'version': '1.2.4'},
      ],
     '1.2.3',
     {'version': '1.2.3'}),
    ([{'version': '1.2.2'},
      {'version': '1.2.3'},
      {'version': '1.2.4'},
      ],
     '1.2.4',
     {'version': '1.2.4'}),
    # an example with full repo_version objects
    (EXAMPLE_REPO_VERSIONS_LIST,
     '1.1.0',
     EXAMPLE_REPO_VERSION_LATEST),
    # if repoversions somehow ends up with two repo_version objects with same version
    ([{'version': '1.2.2',
       'url': 'http://thefirsturl.example.com'},
      {'version': '1.2.2',
       'url': 'http://thesecondurl.example.com'},
      ],
     '1.2.2',
     {'version': '1.2.2',
      'url': 'http://thesecondurl.example.com'}),
])
def test_select_collection_version(repoversions, version, expected):
    log.debug('repoversions: %s', repoversions)
    log.debug('version: %s', version)
    log.debug('expected: %s', expected)
    res = galaxy_url.select_repository_version(repoversions, version)

    assert isinstance(res, dict)

    if res and expected:
        assert res['version'] == expected['version']

    assert res == expected


# see https://github.com/ansible/mazer/issues/79
def test_select_collection_version_empty_repoversions():
    repoversions = []
    version = '1.2.3'

    res = galaxy_url.select_repository_version(repoversions, version)

    assert isinstance(res, dict)
    assert res == {}


# def test_get_download_url_no_data():
#    res = galaxy_url.get_download_url()

#    log.debug('res: %s', res)


def test_get_download_url_empty_data():
    repo_data = {}
    external_url = ""
    repoversion = []
    res = galaxy_url.get_download_url(repo_data=repo_data, external_url=external_url, repoversion=repoversion)

    log.debug('res: %s', res)
    assert res is None


def test_get_download_url_download_url_in_repo_data():
    download_url = 'https://github.com/alikins/collection_reqs_test/archive/0.0.13.tar.gz'
    repo_data = {'download_url': download_url}
    external_url = 'https://github.com/alikins/collection_reqs_test'
    repoversion = []
    res = galaxy_url.get_download_url(repo_data=repo_data, external_url=external_url, repoversion=repoversion)

    log.debug('res: %s', res)
    assert res == download_url


def test_get_download_url_download_url_in_repo_version():
    repo_data = {}
    external_url = 'https://github.com/alikins/collection_reqs_test'
    repoversion = EXAMPLE_REPO_VERSIONS_LIST[0]
    res = galaxy_url.get_download_url(repo_data=repo_data, external_url=external_url, repoversion=repoversion)

    log.debug('res: %s', res)
    assert res == EXAMPLE_REPO_VERSIONS_LIST[0]['download_url']


def test_get_download_url_no_download_url_in_repodata_or_repoversion():
    repo_data = {}
    external_url = 'https://github.com/alikins/collection_reqs_test'
    repoversion = EXAMPLE_REPO_VERSIONS_LIST[0].copy()
    del repoversion['download_url']

    res = galaxy_url.get_download_url(repo_data=repo_data, external_url=external_url, repoversion=repoversion)

    log.debug('res: %s', res)
    assert res == 'https://github.com/alikins/collection_reqs_test/archive/1.0.6.tar.gz'



import logging

import attr
import semantic_version

from ansible_galaxy.utils.version import convert_string_to_version_spec, version_needs_aka, \
    normalize_version_string

log = logging.getLogger(__name__)


@attr.s(frozen=True)
class RequirementSpec(object):
    '''The info used to identify a requirement.

    ie, the namespace, name, and the spec of what version is
    required'''

    namespace = attr.ib()
    name = attr.ib()
    version_spec = attr.ib(type=semantic_version.Spec, default=semantic_version.Spec('*'),
                           converter=convert_string_to_version_spec)

    # This is for supporting 'v1.0.0' etc. In that case the version_spec is '==1.0.0' and
    # version_aka is 'v1.0.0'. Need to track it because it is used to build github download
    # urls sometimes
    version_aka = attr.ib(default=None)
    fetch_method = attr.ib(default=None, cmp=False)
    src = attr.ib(default=None, cmp=False)
    req_spec_string = attr.ib(default=None, cmp=False)

    @property
    def label(self):
        return '%s.%s' % (self.namespace, self.name)

    @classmethod
    def from_dict(cls, data):
        version_spec_str = data.get('version_spec', None)

        # If we are specifying a version in the spec_data, then
        # assume we want to install exactly that version, so build
        # a version_spec to indicate that (ie, '==1.0.0' etc)
        if not version_spec_str:
            if data.get('version', None):
                ver = data['version']

                # try to handle matching 'v1.0.0' etc
                if version_needs_aka(str(ver)):
                    data['version'] = normalize_version_string(ver)
                    data['version_aka'] = ver
                version_spec_str = '==%s' % data['version']
            else:
                # No version_spec, and version is None, that means match anything
                version_spec_str = '*'

        instance = cls(namespace=data['namespace'],
                       name=data['name'],
                       version_spec=version_spec_str,
                       version_aka=data.get('version_aka', None),
                       fetch_method=data.get('fetch_method', None),
                       req_spec_string=data.get('req_spec_string', None),
                       src=data.get('src', None),
                       )
        return instance



import logging
import os

from ansible_galaxy import repository
from ansible_galaxy import matchers
from ansible_galaxy import installed_namespaces_db

log = logging.getLogger(__name__)


def get_repository_paths(namespace_path):
    # TODO: abstract this a bit?  one to make it easier to mock, but also
    #       possibly to prepare for nested dirs, multiple paths, various
    #       filters/whitelist/blacklist/excludes, caching, or respecting
    #       fs ordering, etc
    #
    # TODO: do some caching, invalidation based on dir mtimes?
    try:
        # TODO: filter on any rules for what a namespace path looks like
        #       may one being 'somenamespace.somename' (a dot sep ns and name)
        #
        repository_paths = os.listdir(namespace_path)
    except OSError as e:
        log.exception(e)
        log.warning('The namespace path %s did not exist so no repositories were found.',
                    namespace_path)
        repository_paths = []

    return repository_paths


def installed_repository_iterator(galaxy_context,
                                  namespace_match_filter=None,
                                  repository_spec_match_filter=None,
                                  requirement_spec_match_filter=None):
    '''For each repository in galaxy_context.content_path, yield matching repositories'''

    namespace_match_filter = namespace_match_filter or matchers.MatchAll()
    repository_spec_match_filter = repository_spec_match_filter or matchers.MatchAll()
    requirement_spec_match_filter = requirement_spec_match_filter or matchers.MatchAll()

    installed_namespace_db = installed_namespaces_db.InstalledNamespaceDatabase(galaxy_context)

    # TODO: iterate/filter per namespace, then per repository, then per collection/role/etc
    for namespace in installed_namespace_db.select(namespace_match_filter=namespace_match_filter):
        log.debug('Looking for repos in namespace "%s"', namespace.namespace)

        # TODO: filter potential repository_paths based on repository_match_filer before we
        #       try to listdir() instead of after
        repository_paths = get_repository_paths(namespace.path)

        for repository_path in repository_paths:

            # TODO: if we need to distinquish repo from collection or role, we could do it here
            repository_ = repository.load_from_dir(galaxy_context.content_path,
                                                   namespace=namespace.namespace,
                                                   name=repository_path,
                                                   installed=True)

            # log.debug('candidate installed repo (pre filter): %s', repository_)

            if repository_spec_match_filter(repository_):
                log.debug('Found repository "%s" (%s) in namespace "%s"', repository_path, str(repository), namespace.namespace)

                if requirement_spec_match_filter(repository_):
                    log.debug('Found repository "%s" in namespace "%s"',
                              repository_path, namespace.namespace)
                    yield repository_


# TODO: add a get(namespace_id, repository_id) for loading a known ns.n directly without iterating
# TODO: add a contains(namespace_id, repository_id, matchers) for checking for existince
#       without loading the Repository from disk. Useful for things
#       like 'is some_repo_spec already installed?'
class InstalledRepositoryDatabase(object):

    def __init__(self, installed_context=None):
        self.installed_context = installed_context

    # TODO: add a repository_type_filter (ie, 'collection' or 'role' or 'other' etc)
    # TODO: something like namespace_condition or namespace_callable might be more accurate
    # TODO: "search" would be more accurate name for select()
    def select(self, namespace_match_filter=None,
               repository_spec_match_filter=None,
               requirement_spec_match_filter=None):
        # ie, default to select * more or less
        namespace_match_filter = namespace_match_filter or matchers.MatchAll()
        repository_spec_match_filter = repository_spec_match_filter or matchers.MatchAll()
        requirement_spec_match_filter = requirement_spec_match_filter or matchers.MatchAll()

        # log.debug('repository_spec_match_filter: %s', repository_spec_match_filter)

        installed_repositories = installed_repository_iterator(self.installed_context,
                                                               namespace_match_filter=namespace_match_filter,
                                                               repository_spec_match_filter=repository_spec_match_filter,
                                                               requirement_spec_match_filter=requirement_spec_match_filter)

        for matched_installed_repository in installed_repositories:
            yield matched_installed_repository

    def by_repository_spec(self, repository_spec):
        repository_spec_match_filter = matchers.MatchRepositorySpec([repository_spec])
        return self.select(repository_spec_match_filter=repository_spec_match_filter)

    def by_requirement(self, requirement):
        requirement_spec = requirement.requirement_spec

        return self.by_requirement_spec(requirement_spec=requirement_spec)

    def by_requirement_spec(self, requirement_spec):
        requirement_spec_match_filter = matchers.MatchRepositoryToRequirementSpec([requirement_spec])

        return self.select(requirement_spec_match_filter=requirement_spec_match_filter)




import logging

import pytest
import six

from ansible_galaxy.models import galaxy_namespace

log = logging.getLogger(__name__)


def test_galaxy_namespace_no_namespace():
    with pytest.raises(TypeError) as exc_info:
        galaxy_namespace.GalaxyNamespace()

    # different TypeError strings for py2 vs py3
    log.debug('exc_info.getrepr(): %s', exc_info.getrepr(showlocals=True, style='native'))

    if six.PY3:
        # example: TypeError: __init__() missing 1 required positional argument: 'namespace'
        assert exc_info.match("__init__.*missing.*1.*'namespace'")

    if six.PY2:
        # TypeError: __init__() takes at least 2 arguments (1 given)
        assert exc_info.match("__init__.*takes at least.*arguments.*given")


def test_galaxy_namespace_no_path():
    namespace = 'my_namespace'

    gn = galaxy_namespace.GalaxyNamespace(namespace)
    log.debug('gn: %s', gn)

    assert gn.namespace == namespace
    assert gn.path is None


def test_galaxy_namespace():
    namespace = 'my_namespace'
    path = '/dev/null/path_used_in_test_collection_namespace'

    gn = galaxy_namespace.GalaxyNamespace(namespace, path=path)
    log.debug('gn: %s', gn)

    assert gn.namespace == namespace
    assert gn.path == path



import datetime
import logging
import os
import tarfile

from ansible_galaxy import archive
from ansible_galaxy import collection_members
from ansible_galaxy import exceptions
from ansible_galaxy import install_info
from ansible_galaxy.models import content
from ansible_galaxy.models.repository_archive import RepositoryArchiveInfo, CollectionRepositoryArchive
from ansible_galaxy.models.repository_archive import CollectionRepositoryArtifactArchive, TraditionalRoleRepositoryArchive
from ansible_galaxy.models.repository_spec import FetchMethods
from ansible_galaxy.models.install_info import InstallInfo
from ansible_galaxy.models.installation_results import InstallationResults

log = logging.getLogger(__name__)

# TODO: better place to define?
META_MAIN = os.path.join('meta', 'main.yml')
GALAXY_FILE = 'ansible-galaxy.yml'
APB_YAML = 'apb.yml'


def null_display_callback(*args, **kwargs):
    log.debug('display_callback: %s', args)


def extract(repository_spec,
            repository_archive_info,
            content_path,
            extract_archive_to_dir,
            tar_file,
            force_overwrite=False,
            display_callback=None):

    all_installed_paths = []

    # TODO: move to content info validate step in install states?
    if not repository_spec.namespace:
        # TODO: better error
        raise exceptions.GalaxyError('While installing a role , no namespace was found. Try providing one with --namespace')

    # label = "%s.%s" % (repository_namespace, repository_name)

    # 'extract_to_path' is for ex, ~/.ansible/content
    log.debug('About to extract %s "%s" to %s', repository_archive_info.archive_type,
              repository_spec.label, content_path)
    # display_callback('- extracting %s repository from "%s"' % (repository_archive_info.archive_type,
    #                                                           repository_spec.label))

    tar_members = tar_file.members

    # self.log.debug('content_dest_root_subpath: %s', content_dest_root_subpath)

    # self.log.debug('content_dest_root_path1: |%s|', content_dest_root_path)

    # TODO: need to support deleting all content in the dirs we are targetting
    #       first (and/or delete the top dir) so that we clean up any files not
    #       part of the content. At the moment, this will add or update the files
    #       that are in the archive, but it will not delete files on the fs that are
    #       not in the archive
    files_to_extract = []
    for member in tar_members:
        rel_path = member.name

        extract_to_filename_path = os.path.join(extract_archive_to_dir, rel_path)

        # self.log.debug('content_dest_root_path: %s', content_dest_root_path)
        # self.log.debug('content_dest_root_rel_path: %s', content_dest_root_rel_path)

        files_to_extract.append({
            'archive_member': member,
            # Note: for trad roles, we are extract the top level of the archive into
            #       a sub path of the destination
            'dest_dir': extract_archive_to_dir,
            'dest_filename': extract_to_filename_path,
            'force_overwrite': force_overwrite})

    file_extractor = archive.extract_files(tar_file, files_to_extract)

    installed_paths = [x for x in file_extractor]

    all_installed_paths.extend(installed_paths)

    log.debug('Extracted %s files from %s %s to %s',
              len(all_installed_paths),
              repository_archive_info.archive_type,
              repository_spec.label,
              extract_archive_to_dir)

    # TODO: InstallResults object? installedPaths, InstallInfo, etc?
    return all_installed_paths


def detect_repository_archive_type(archive_path, file_names):
    '''Try to determine if we are a role, multi-content, apb etc.

    if there is a meta/main.yml ->  role

    if there is any of the content types subdirs -> multi-content'''

    # FIXME: just looking for the root dir...

    top_dir = file_names[0]

    # log.debug('top_dir of %s: %s', archive_path, top_dir)

    meta_main_target = os.path.join(top_dir, 'meta/main.yml')
    manifest_path = os.path.join(top_dir, 'MANIFEST.json')

    type_dirs = content.CONTENT_TYPE_DIR_MAP.values()
    # log.debug('type_dirs: %s', type_dirs)

    type_dir_targets = set([os.path.join(top_dir, x) for x in type_dirs])
    # log.debug('type_dir_targets: %s', type_dir_targets)

    has_manifest = any([member for member in file_names if member == manifest_path])
    log.debug('has_manifest: %s', has_manifest)

    for member in file_names:
        if member == meta_main_target:
            return 'role'

    for member in file_names:
        if member in type_dir_targets:
            if has_manifest:
                return 'multi-content-artifact'
            return 'multi-content'

    # otherwise, assume it is a collection / multi-content repo archive
    return 'multi-content'


def build_archive_info(archive_path, file_names):
    archive_parent_dir = None
    archive_parent_dir = file_names[0]

    archive_type = detect_repository_archive_type(archive_path, file_names)

    # log.debug('archive_type of %s: %s', archive_path, archive_type)
    # log.debug("archive_parent_dir of %s: %s", archive_path, archive_parent_dir)

    # looks like we are a role, update the default content_type from all -> role
    if archive_type == 'role':
        log.debug('Found role metadata in the archive %s, so installing it as role content_type',
                  archive_path)

    archive_info = RepositoryArchiveInfo(top_dir=archive_parent_dir,
                                         archive_type=archive_type,
                                         archive_path=archive_path)

    return archive_info


def load_editable_archive_info(archive_path, repository_spec):
    file_walker = collection_members.FileWalker(collection_path=archive_path)
    col_members = collection_members.CollectionMembers(walker=file_walker)
    file_members = list(col_members.run())

    archive_info = build_archive_info(archive_path, file_members)

    return archive_info, None


def load_tarfile_archive_info(archive_path, repository_spec):

    # if not tarfile.is_tarfile(archive_path):
    #    raise exceptions.GalaxyClientError("the file downloaded was not a tar.gz")

    if archive_path.endswith('.gz'):
        tar_flags = "r:gz"
    else:
        tar_flags = "r"

    try:
        repository_tar_file = tarfile.open(archive_path, tar_flags)
    except tarfile.TarError as e:
        log.exception(e)
        raise exceptions.GalaxyClientError("Error opening the tar file %s with flags: %s for repo: %s" %
                                           (archive_path, tar_flags, repository_spec))

    members = repository_tar_file.getmembers()
    archive_info = build_archive_info(archive_path, [m.name for m in members])

    # log.debug('role archive_info for %s: %s', archive_path, archive_info)

    return archive_info, repository_tar_file


def load_archive_info(archive_path, repository_spec=None):

    # "installing" an existing dir as editable
    if repository_spec and repository_spec.fetch_method == FetchMethods.EDITABLE:
        return load_editable_archive_info(archive_path, repository_spec)

    return load_tarfile_archive_info(archive_path, repository_spec)


def load_archive(archive_path, repository_spec=None):
    '''Load the archive file at archive_path and return a RepositoryArchive'''
    # To avoid opening the archive file twice, and since we have to open/load it to
    # get the archive_info, we also return it from load_archive_info
    archive_info, tar_file = load_archive_info(archive_path, repository_spec)

    # factory-ish
    if archive_info.archive_type in ['role']:
        repository_archive_ = TraditionalRoleRepositoryArchive(info=archive_info,
                                                               tar_file=tar_file)
    elif archive_info.archive_type == 'multi-content-artifact':
        repository_archive_ = CollectionRepositoryArtifactArchive(info=archive_info,
                                                                  tar_file=tar_file)
    else:
        repository_archive_ = CollectionRepositoryArchive(info=archive_info,
                                                          tar_file=tar_file)

    log.debug('repository archive_ for %s: %s', archive_path, repository_archive_)

    return repository_archive_


def install(repository_archive, repository_spec, destination_info, display_callback):
    log.debug('installing/extracting repo archive %s to destination %s', repository_archive, destination_info)

    # An editable install is a symlink to existing dir, so nothing to extract
    if destination_info.editable:
        all_installed_files = []
    else:
        all_installed_files = extract(repository_spec,
                                      repository_archive.info,
                                      content_path=destination_info.destination_root_dir,
                                      extract_archive_to_dir=destination_info.extract_archive_to_dir,
                                      tar_file=repository_archive.tar_file,
                                      display_callback=display_callback)

    install_datetime = datetime.datetime.utcnow()

    install_info_ = InstallInfo.from_version_date(repository_spec.version,
                                                  install_datetime=install_datetime)

    # TODO: this save will need to be moved to a step later. after validating install?
    # The to_dict_version_strings is to convert the un-yaml-able semantic_version.Version to a string
    install_info.save(install_info_.to_dict_version_strings(),
                      destination_info.install_info_path)

    installation_results = InstallationResults(install_info_path=destination_info.install_info_path,
                                               install_info=install_info_,
                                               installed_to_path=destination_info.path,
                                               installed_datetime=install_datetime,
                                               installed_files=all_installed_files)
    return installation_results



import collections
import json
import logging
import os

from ansible_galaxy import repository

log = logging.getLogger(__name__)


def _load_path(file_name):
    # test_data_path = os.path.join(os.path.dirname(__file__), '../../data/%s' % file_name)
    test_data_path = os.path.join(os.path.dirname(__file__), '../data/%s' % file_name)

    log.debug('test_data_path=%s', test_data_path)

    data = open(test_data_path, 'r').read()
    ret = repository.load(data)
    return ret


def test_load_sample():
    ds = _load_path('sample-ansible-galaxy.yml')

    log.debug('ds=%s', json.dumps(ds, indent=4))

    assert isinstance(ds, collections.Mapping)


def test_load_sample2():
    ds = _load_path('sample-ansible-galaxy2.yml')

    log.debug('ds=%s', json.dumps(ds, indent=4))

    assert isinstance(ds, collections.Mapping)



import logging

import semantic_version

from ansible_galaxy import exceptions
from ansible_galaxy.utils.version import normalize_version_string

log = logging.getLogger(__name__)


def sort_versions(versions):
    # list of tuples of loose_version and original string, the sort
    # will sort based on first value of tuple, then we return just the
    # original strings
    semver_versions = [(semantic_version.Version(a), a) for a in versions]
    semver_versions.sort()
    return [v[1] for v in semver_versions]


# TODO: somewhere we need to have a code path for the ordered version list returned from server
#       and another for the versions we got elsewhere.
def get_latest_version(available_normalized_versions, content_data):
    # and sort them to get the latest version. If there
    # are no versions in the list, we'll grab the head
    # of the master branch
    if available_normalized_versions:
        try:
            sorted_versions = sort_versions(available_normalized_versions)
        except (TypeError, ValueError) as e:
            log.exception(e)
            raise exceptions.GalaxyClientError(
                'Unable to compare content versions (%s) to determine the most recent version due to incompatible version formats. '
                'Please contact the content author to resolve versioning conflicts, or specify an explicit content version to install. '
                'The version error was: "%s"' % (', '.join(available_normalized_versions), e)
            )

        content_version = sorted_versions[-1]
    # FIXME: follow 'repository' branch and it's ['import_branch'] ?
    elif content_data.get('github_branch', None):
        content_version = content_data['github_branch']
    else:
        content_version = 'master'

    return content_version


def normalize_versions(content_versions):
    # a list of tuples of (normalized_version, original_version) for building
    # map of normalized version to original version
    normalized_versions = [(normalize_version_string(x), x) for x in content_versions]
    available_normalized_versions = [v[0] for v in normalized_versions]

    # map the 'normalized' version back to the original version string, we need it for
    # content archive download urls
    # norm_to_orig_map = dict(normalized_versions)
    norm_to_orig_map = normalized_versions

    return (available_normalized_versions, norm_to_orig_map)


def validate_versions(content_versions):
    valid_versions = []
    invalid_versions = []
    for version in content_versions:
        if not semantic_version.validate(version):
            log.warning('The version string "%s" is not valid, skipping.', version)
            invalid_versions.append(version)
            continue
        valid_versions.append(version)

    return (valid_versions, invalid_versions)


def get_repository_version(repository_data, requirement_spec, repository_versions):
    '''find and compare repository version found in repository_data dict

    repository_data is a dict based on /api/v1/repositories/13 for ex
    content_content_data is the name of the content specified by user?
    version is the version string asked for by user
    content_versions is a list of version strings in order
    '''

    log.debug('%s wants ver: %s type: %s', requirement_spec.name, requirement_spec.version_spec, type(requirement_spec.version_spec))

    # normalize versions, but also build a map of the normalized version string to the orig string
    available_normalized_versions, norm_to_orig_map = normalize_versions(repository_versions)

    # verify that the normalized versions are valid semver now so that we dont worry about it
    # in the sort
    available_versions, dummy = \
        validate_versions(available_normalized_versions)

    # we specified a particular version is required so look for it in available versions
    if not available_versions:
        # FIXME: should we show the actual available versions or the available
        #        versions we searched in?  act: ['v1.0.0', '1.1'] nor: ['1.0.0', '1.1']
        msg = "- The list of available versions for %s is empty (%s)." % \
            (requirement_spec.label or 'content', available_versions)
        raise exceptions.GalaxyError(msg)

    semver_available_versions = [semantic_version.Version(ver) for ver in available_versions]

    latest_version = requirement_spec.version_spec.select(semver_available_versions)
    if latest_version is None:
        # TODO: how do we msg 'couldn't find the version you specified
        #       in actual version tags or ones we made up without the leading v'
        msg = "- the specified version spec (%s) of %s was not found in the list of available versions (%s)." % \
            (requirement_spec.version_spec, requirement_spec.label or 'content', available_versions)
        raise exceptions.GalaxyError(msg)

    # if we get here, 'version' is in available_normalized_versions
    # return the exact match version since it was available
    latest_version_str = str(latest_version)

    # have to use list of tuples here since we can have multiple identical keys with different values
    norm_to_orig = [x for x in norm_to_orig_map if x[0] == latest_version_str]

    if len(norm_to_orig) > 1:
        unnormal_versions = [ver_alias[1] for ver_alias in norm_to_orig]
        raise exceptions.GalaxyClientError('There are ambiguous and contradicting version numbers (%s) that match version spec "%s"' %
                                           (', '.join(unnormal_versions), str(requirement_spec.version_spec)))

    orig_version_str = norm_to_orig[0][1]

    log.debug('%s requested ver: %s, matched: %s, using real ver: %s ', requirement_spec.label, requirement_spec.version_spec, latest_version, orig_version_str)

    return orig_version_str



import logging
import tempfile

from ansible_galaxy import exceptions
from ansible_galaxy.flat_rest_api.urls import open_url

log = logging.getLogger(__name__)


# FIXME: let the archive_url be passed in
def fetch_url(archive_url, validate_certs=True):
    """
    Downloads the archived content from github to a temp location
    """

    # TODO: should probably be based on/shared with rest API client code, so that
    #       content downloads could support any thing the rest code does
    #       (ie, any TLS cert setup, proxy config, auth options, etc)
    # WHEN: if we change the underlying http client impl at least
    try:
        url_file = open_url(archive_url, validate_certs=validate_certs)

        temp_file = tempfile.NamedTemporaryFile(delete=False,
                                                prefix='tmp-ansible-galaxy-content-archive-',
                                                suffix='.tar.gz')

        data = url_file.read()
        while data:
            temp_file.write(data)
            data = url_file.read()
        temp_file.close()
        return temp_file.name
    except Exception as e:
        # FIXME: there is a ton of reasons a download and save could fail so could likely provided better errors here
        log.exception(e)
        raise exceptions.GalaxyDownloadError(e, url=archive_url)

    return False



import logging
import os

from ansible_galaxy import exceptions

log = logging.getLogger(__name__)


class EditableFetch(object):
    fetch_method = 'editable'

    def __init__(self, galaxy_context, requirement_spec):
        self.galaxy_context = galaxy_context
        self.requirement_spec = requirement_spec
        self.local_path = self.requirement_spec.src

    def find(self):
        real_path = os.path.abspath(self.requirement_spec.src)

        log.debug('Searching for a repository to link to as an editable install at %s', real_path)

        if not os.path.isdir(real_path):
            log.warning("%s needs to be a local directory for an editable install" % self.repository.src)
            raise exceptions.GalaxyClientError('Error finding an editable install of %s because %s is not a directory', self.requirement_spec.src, real_path)

        results = {'content': {'galaxy_namespace': self.requirement_spec.namespace,
                               'repo_name': self.requirement_spec.name},
                   'custom': {'real_path': self.requirement_spec.src}
                   }

        return results

    def fetch(self, find_results=None):
        find_results = find_results or {}

        real_path = find_results.get('custom', {}).get('real_path', None)
        if not real_path:
            raise exceptions.GalaxyClientError('Error fetching an editable install of %s because no "real_path" was found in find_results',
                                               self.requirement_spec.src, real_path)

        dst_ns_root = os.path.join(self.galaxy_context.content_path, self.requirement_spec.namespace)
        dst_repo_root = os.path.join(dst_ns_root,
                                     self.requirement_spec.name)

        if not os.path.exists(dst_ns_root):
            os.makedirs(dst_ns_root)

        if not os.path.exists(dst_repo_root):
            os.symlink(real_path, dst_repo_root)

        repository_archive_path = self.local_path

        log.debug('repository_archive_path=%s (inplace) synlink to %s',
                  repository_archive_path,
                  real_path)

        results = {'archive_path': repository_archive_path,
                   'fetch_method': self.fetch_method}

        results['custom'] = {'local_path': self.local_path,
                             'real_path': real_path,
                             'symlinked_repo_root': dst_repo_root}
        results['content'] = find_results['content']

        return results

    def cleanup(self):
        return None



"""Setup default logging"""

import logging
import logging.config
import os
import yaml

LOG_FILE = os.path.expandvars(os.path.expanduser('~/.ansible/mazer.log'))

DEFAULT_CONSOLE_LEVEL = os.getenv('MAZER_LOG_LEVEL', 'WARNING').upper()
DEFAULT_LEVEL = 'DEBUG'

DEFAULT_LOGGING_CONFIG_YAML = os.path.join(os.path.dirname(__file__), 'default-mazer-logging.yml')
LOGGING_CONFIG_YAML = os.path.expandvars(os.path.expanduser('~/.ansible/mazer-logging.yml'))

FALLBACK_LOGGING_CONFIG = {
    'version': 1,

    'disable_existing_loggers': False,

    'handlers': {
        'null_handler': {
            'class': 'logging.NullHandler',
            'level': 'ERROR',
        },
    },

    'loggers': {
        'ansible_galaxy': {
            'handlers': ['null_handler'],
            'level': 'INFO',
        },
        'ansible_galaxy_cli': {
            'handlers': ['null_handler'],
            'level': 'INFO',
        },
    }
}


class ExpandTildeWatchedFileHandler(logging.handlers.WatchedFileHandler):
    '''A variant of WatchedFileHandler that will expand ~/ in it's filename param'''
    def __init__(self, *args, **kwargs):
        orig_filename = kwargs.pop('filename', '~/.ansible/mazer.log')
        kwargs['filename'] = os.path.expandvars(os.path.expanduser(orig_filename))
        logdir = os.path.dirname(kwargs['filename'])
        if not os.path.exists(logdir):
            os.makedirs(logdir)
        super(ExpandTildeWatchedFileHandler, self).__init__(*args, **kwargs)


def setup(logging_config=None):
    logging_config = logging_config or {}

    logging.config.dictConfig(logging_config)


def load_config_yaml(config_file_path):
    logging_config = None

    try:
        with open(config_file_path, 'r') as logging_config_file:
            logging_config = yaml.safe_load(logging_config_file)
    except (IOError, OSError):
        pass
    except yaml.error.YAMLError as e:
        print(e)

    return logging_config


def setup_default():
    logging_config = None

    # fallback is basically no setup, null handler, etc
    # builtin, doesn't depend on yaml config
    setup(FALLBACK_LOGGING_CONFIG)

    # load the more extensive defaults from DEFAULT_LOGGING_CONFIG_YAML
    default_logging_config = load_config_yaml(DEFAULT_LOGGING_CONFIG_YAML)
    if default_logging_config:
        setup(default_logging_config)

    # load custom logging config
    logging_config = load_config_yaml(LOGGING_CONFIG_YAML)

    if logging_config:
        setup(logging_config)

    # import logging_tree
    # logging_tree.printout()



# (c) 2016, Adrian Likins <alikins@redhat.com>
#
# This file is original part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.

# Make coding more python3-ish
from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

# import ansible
import errno
import os
import logging
import shutil
import tarfile
import tempfile

try:
    import unittest2 as unittest
except ImportError:
    import unittest

from mock import call, patch


# FIXME: shouldn't need to patch object directly
import ansible_galaxy_cli
import ansible_galaxy

from ansible_galaxy import exceptions

from ansible_galaxy_cli.cli.galaxy import GalaxyCLI
from ansible_galaxy_cli import exceptions as cli_exceptions

log = logging.getLogger(__name__)


def ensure_dir(path):
    try:
        os.makedirs(path)
    except OSError as e:
        if e.errno != errno.EEXIST:
            log.exception(e)
            raise


class TestGalaxy(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        '''creating prerequisites for installing a role; setUpClass occurs ONCE whereas setUp occurs with every method tested.'''
        # class data for easy viewing: role_dir, role_tar, role_name, role_req, role_path

        if os.path.exists("./delete_me"):
            shutil.rmtree("./delete_me")

        archive_path = os.path.dirname(os.path.abspath(__file__))
        archive_path = os.path.join(archive_path, 'data', 'delete_me.tar.gz')
        tar = tarfile.open(archive_path, 'r:gz')
        tar.extractall(path=".")
        cls.role_dir = "./delete_me"
        cls.role_name = "unittest.delete_me"

        # add a meta/main.yml
        test_dir = os.path.join(cls.role_dir, 'meta')
        ensure_dir(test_dir)
        with open(os.path.join(cls.role_dir, 'meta/main.yml'), 'w') as fd:
            fd.write('galaxy_info: {}\ndependencies: {}')

        # making a temp dir for role installation
        cls.role_path = os.path.join(tempfile.mkdtemp(), "roles")

        ensure_dir(cls.role_path)

        # creating a tar file name for class data
        cls.role_tar = './delete_me.tar.gz'
        cls.makeTar(cls.role_tar, cls.role_dir)

        # creating a temp file with installation requirements
        cls.role_req = './delete_me_requirements.yml'
        dep_lines = ["- 'src': '%s'\n" % cls.role_tar,
                     "  'name': '%s'\n" % cls.role_name,
                     "  'path': '%s'\n" % cls.role_path]
        with open(cls.role_req, 'w') as fd:
            for dep_line in dep_lines:
                fd.write(dep_line)

    @classmethod
    def makeTar(cls, output_file, source_dir):
        ''' used for making a tarfile from a role directory '''
        # adding directory into a tar file
        try:
            tar = tarfile.open(output_file, "w:gz")
            tar.add(source_dir, arcname=os.path.basename(source_dir))
        except AttributeError:  # tarfile obj. has no attribute __exit__ prior to python 2.7
            pass
        finally:  # ensuring closure of tarfile obj
            tar.close()

    @classmethod
    def tearDownClass(cls):
        '''After tests are finished removes things created in setUpClass'''
        # deleting the temp role directory
        if os.path.exists(cls.role_dir):
            shutil.rmtree(cls.role_dir)
        if os.path.exists(cls.role_req):
            os.remove(cls.role_req)
        if os.path.exists(cls.role_tar):
            os.remove(cls.role_tar)
        if os.path.isdir(cls.role_path):
            shutil.rmtree(cls.role_path)

    def setUp(self):
        self.default_args = ['ansible-galaxy']

    def test_run(self):
        ''' verifies that the GalaxyCLI object's api is created and that execute() is called. '''
        gc = GalaxyCLI(args=["ansible-galaxy", "install", "--ignore-errors", "imaginary_role"])
        gc.parse()
        with patch.object(ansible_galaxy_cli.cli.CLI, "execute", return_value=None) as mock_ex:
            with patch.object(ansible_galaxy_cli.cli.CLI, "run", return_value=None) as mock_run:
                gc.run()

                # testing
                self.assertEqual(mock_run.call_count, 1)
                self.assertTrue(isinstance(gc.api, ansible_galaxy.rest_api.GalaxyAPI))
                self.assertEqual(mock_ex.call_count, 1)

    def test_execute_remove(self):
        # installing role
        log.debug('self.role_path: %s', self.role_path)
        log.debug('self.role_req: %s', self.role_req)

        gc = GalaxyCLI(args=["ansible-galaxy", "install", "--content-path", self.role_path, "-r", self.role_req, '--force'])
        gc.parse()
        gc.run()

        log.debug('self.role_name: %s', self.role_name)
        # location where the role was installed
        role_file = os.path.join(self.role_path, self.role_name)

        log.debug('role_file: %s', role_file)

        # removing role
        # args = ["ansible-galaxy", "remove", role_file, self.role_name]
        args = ["ansible-galaxy", "remove", self.role_name]
        log.debug('args: %s', args)
        gc = GalaxyCLI(args=args)
        gc.parse()
        gc.run()

        # testing role was removed
        removed_role = not os.path.exists(role_file)
        self.assertTrue(removed_role)

    def test_exit_without_ignore_without_flag(self):
        ''' tests that GalaxyCLI exits with the error specified if the --ignore-errors flag is not used '''
        gc = GalaxyCLI(args=["ansible-galaxy", "install", "--server=None", "testing.fake_role_name"])
        gc.parse()
        # testing that error expected is raised
        self.assertRaises(exceptions.GalaxyError, gc.run)
        # self.assertTrue(mocked_display.called_once_with("- downloading role 'fake_role_name', owned by "))

    def test_exit_without_ignore_with_flag(self):
        ''' tests that GalaxyCLI exits without the error specified if the --ignore-errors flag is used  '''
        # testing with --ignore-errors flag
        gc = GalaxyCLI(args=["ansible-galaxy", "install", "--server=None", "testing.fake_role_name", "--ignore-errors"])
        gc.parse()
        gc.run()
        #    self.assertTrue(mocked_display.called_once_with("- downloading role 'fake_role_name', owned by "))

    def run_parse_common(self, galaxycli_obj, action):
        with patch.object(ansible_galaxy_cli.cli.SortedOptParser, "set_usage") as mocked_usage:
            galaxycli_obj.parse()

            # checking that the common results of parse() for all possible actions have been created/called
            self.assertIsInstance(galaxycli_obj.parser, ansible_galaxy_cli.cli.SortedOptParser)
            # self.assertIsInstance(galaxycli_obj.galaxy, ansible_galaxy.models.context.GalaxyContext)
            formatted_call = {
                'info': 'usage: %prog info [options] repo_name[,version]',
                'install': 'usage: %prog install [options] [-r FILE | repo_name(s)[,version] | scm+repo_url[,version] | tar_file(s)]',
                'list': 'usage: %prog list [repo_name]',
                'publish': 'usage: %prog publish [options] archive_path',
                'remove': 'usage: %prog remove repo1 repo2 ...',
                'version': 'usage: %prog version',
            }

            first_call = 'usage: %prog [build|info|install|list|migrate_role|publish|remove|version] [--help] [options] ...'
            second_call = formatted_call[action]
            calls = [call(first_call), call(second_call)]
            mocked_usage.assert_has_calls(calls)

    def test_parse_no_action(self):
        ''' testing the options parser when no action is given '''
        gc = GalaxyCLI(args=["ansible-galaxy", ""])
        self.assertRaises(cli_exceptions.CliOptionsError, gc.parse)

    def test_parse_invalid_action(self):
        ''' testing the options parser when an invalid action is given '''
        gc = GalaxyCLI(args=["ansible-galaxy", "NOT_ACTION"])
        self.assertRaises(cli_exceptions.CliOptionsError, gc.parse)

    def test_parse_info(self):
        ''' testing the options parser when the action 'info' is given '''
        gc = GalaxyCLI(args=["ansible-galaxy", "info"])
        self.run_parse_common(gc, "info")
        self.assertEqual(gc.options.offline, False)

    def test_parse_install(self):
        ''' testing the options parser when the action 'install' is given '''
        gc = GalaxyCLI(args=["ansible-galaxy", "install"])
        self.run_parse_common(gc, "install")
        self.assertEqual(gc.options.ignore_errors, False)
        self.assertEqual(gc.options.no_deps, False)
        self.assertEqual(gc.options.role_file, None)
        self.assertEqual(gc.options.force, False)

    def test_parse_list(self):
        ''' testing the options parser when the action 'list' is given '''
        gc = GalaxyCLI(args=["ansible-galaxy", "list"])
        self.run_parse_common(gc, "list")
        self.assertEqual(gc.options.verbosity, 0)

    def test_parse_publish(self):
        ''' testing the options parser when the action 'publish' is given '''
        gc = GalaxyCLI(args=["ansible-galaxy", "publish"])
        self.run_parse_common(gc, "publish")
        self.assertEqual(gc.options.verbosity, 0)

    def test_parse_remove(self):
        ''' testing the options parser when the action 'remove' is given '''
        gc = GalaxyCLI(args=["ansible-galaxy", "remove"])
        self.run_parse_common(gc, "remove")
        self.assertEqual(gc.options.verbosity, 0)

    def test_parse_version(self):
        ''' testing the options parser when the action 'version' is given '''
        gc = GalaxyCLI(args=["ansible-galaxy", "version"])
        self.run_parse_common(gc, "version")
        self.assertEqual(gc.options.verbosity, 0)



# strategy for installing a Collection (name resolve it, find it,
#  fetch it's artifact, validate/verify it, install/extract it, update install dbs, etc)
import logging
import os
import pprint

import attr

from ansible_galaxy import repository_archive
from ansible_galaxy import exceptions
from ansible_galaxy import installed_repository_db
from ansible_galaxy.models.install_destination import InstallDestinationInfo
from ansible_galaxy.models.repository_spec import FetchMethods, RepositorySpec

log = logging.getLogger(__name__)

# This should probably be a state machine for stepping through the install states
# See actions.install.install_collection for a sketch of the states


def find(fetcher):
    """find/discover info about the content"""

    find_results = fetcher.find()

    return find_results


# def fetch(fetcher, collection):
#    pass

def fetch(fetcher, repository_spec, find_results):
    """download the archive and side effect set self._archive_path to where it was downloaded to.

    MUST be called after self.find()."""

    log.debug('Fetching repository_spec=%s', repository_spec)

    try:
        # FIXME: note that ignore_certs for the galaxy
        # server(galaxy_context.server['ignore_certs'])
        # does not really imply that the repo archive download should ignore certs as well
        # (galaxy api server vs cdn) but for now, we use the value for both
        fetch_results = fetcher.fetch(find_results=find_results)
    except exceptions.GalaxyDownloadError as e:
        log.exception(e)

        # TODO: having to keep fetcher state for tracking fetcher.remote_resource and/or cleanup
        #       is kind of annoying.These methods may need to be in a class. Or maybe
        #       the GalaxyDownloadError shoud/could have any info.
        blurb = 'Failed to fetch the content archive "%s": %s'
        log.error(blurb, fetcher.remote_resource, e)

        # reraise, currently handled in main
        # TODO: if we support more than one archive per invocation, will need to accumulate errors
        #       if we want to skip some of them
        raise

    return fetch_results

#
# The caller of install() may be the best place to figure out things like where to
# extract the content too. Can likely handle figuring out if artifact is a collection_artifact
# or a trad_role_artifact and call different things. Or better, create an approriate
# ArchiveArtifact and just call it's extract()/install() etc.
#
# def install(fetch_results, enough_info_to_figure_out_where_to_extract_etc,
#            probably_a_progress_display_callback, maybe_an_error_callback):
#    pass

    # return install_time? an InstallInfo? list of InstalledCollection?


def repository_spec_from_find_results(find_results,
                                      requirement_spec):
    '''Create a new RepositorySpec with updated info from fetch_results.

    Evolves repository_spec to match fetch results.'''
    # TODO: do we still need to check the fetched version against the spec version?
    #       We do, since the unspecific version is None, so fetched versions wont match
    #       so we need a new repository_spec for install.
    # TODO: this is more or less a verify/validate step or state transition
    content_data = find_results.get('content', {})
    resolved_version = content_data.get('version')
    log.debug('version_spec "%s" for %s was requested and was resolved to version "%s"',
              requirement_spec.version_spec, requirement_spec.label,
              resolved_version)

    # repository_spec = attr.evolve(repository_spec, version=content_data['fetched_version'])
    # In theory, a fetch can return a different namespace/name than the one request. This
    # is for things like server side aliases.
    resolved_name = content_data.get('fetched_name', requirement_spec.name)
    resolved_namespace = content_data.get('content_namespace', requirement_spec.namespace)

    # Build a RepositorySpec based on RequirementSpec and the extra info resolved in find()
    spec_data = attr.asdict(requirement_spec)

    del spec_data['version_spec']

    spec_data['version'] = resolved_version
    spec_data['namespace'] = resolved_namespace
    spec_data['name'] = resolved_name

    repository_spec = RepositorySpec.from_dict(spec_data)
    return repository_spec


def install(galaxy_context,
            fetcher,
            fetch_results,
            repository_spec,
            force_overwrite=False,
            display_callback=None):
    """extract the archive to the filesystem and write out install metadata.

    MUST be called after self.fetch()."""

    log.debug('install: repository_spec=%s, force_overwrite=%s',
              repository_spec, force_overwrite)
    just_installed_spec_and_results = []

    # FIXME: really need to move the fetch step elsewhere and do it before,
    #        install should get pass a content_archive (or something more abstract)
    # TODO: some useful exceptions for 'cant find', 'cant read', 'cant write'

    archive_path = fetch_results.get('archive_path', None)

    # TODO: this could be pulled up a layer, after getting fetch_results but before install()
    if not archive_path:
        raise exceptions.GalaxyClientError('No valid content data found for...')

    log.debug("installing from %s", archive_path)

    # TODO: this is figuring out the archive type (multi-content collection or a trad role)
    #       could potentially pull this up a layer
    repo_archive_ = repository_archive.load_archive(archive_path, repository_spec)

    log.debug('repo_archive_: %s', repo_archive_)
    log.debug('repo_archive_.info: %s', repo_archive_.info)

    # we strip off any higher-level directories for all of the files contained within
    # the tar file here. The default is 'github_repo-target'. Gerrit instances, on the other
    # hand, does not have a parent directory at all.

    # preparation for archive extraction
    if not os.path.isdir(galaxy_context.content_path):
        log.debug('No content path (%s) found so creating it', galaxy_context.content_path)

        os.makedirs(galaxy_context.content_path)

    # Build up all the info about where the repository will be installed to
    namespaced_repository_path = '%s/%s' % (repository_spec.namespace,
                                            repository_spec.name)

    install_info_path = os.path.join(galaxy_context.content_path,
                                     namespaced_repository_path,
                                     'meta/.galaxy_install_info')

    # extract_archive_to_dir depends on the repo_archive type, so ask it
    extract_archive_to_dir = os.path.join(galaxy_context.content_path,
                                          namespaced_repository_path,
                                          repo_archive_.repository_dest_root_subpath(repository_spec.name))

    editable = repository_spec.fetch_method == FetchMethods.EDITABLE

    destination_info = InstallDestinationInfo(destination_root_dir=galaxy_context.content_path,
                                              repository_spec=repository_spec,
                                              extract_archive_to_dir=extract_archive_to_dir,
                                              namespaced_repository_path=namespaced_repository_path,
                                              install_info_path=install_info_path,
                                              force_overwrite=force_overwrite,
                                              editable=editable)

    # A list of InstallationResults
    res = repository_archive.install(repo_archive_,
                                     repository_spec=repository_spec,
                                     destination_info=destination_info,
                                     display_callback=display_callback)

    just_installed_spec_and_results.append((repository_spec, res))

    if display_callback:
        display_callback("- The repository %s was successfully installed to %s" % (repository_spec.label,
                                                                                   galaxy_context.content_path))

    # rm any temp files created when getting the content archive
    # TODO: use some sort of callback?
    fetcher.cleanup()

    # We know the repo specs for the repos we asked to install, and the installation results,
    # so now use that info to find the just installed repos on disk and load them and return them.
    just_installed_repository_specs = [x[0] for x in just_installed_spec_and_results]

    # log.debug('just_installed_repository_specs: %s', just_installed_repository_specs)

    irdb = installed_repository_db.InstalledRepositoryDatabase(galaxy_context)

    just_installed_repositories = []

    for just_installed_repository_spec in just_installed_repository_specs:
        just_installed_repository_gen = irdb.by_repository_spec(just_installed_repository_spec)

        # TODO: Eventually, we could make install.install return a generator and yield these results straigt
        #       from just_installed_repository_generator. The loop here is mostly for logging/feedback.
        # Should only get one answer here for now.
        for just_installed_repository in just_installed_repository_gen:
            log.debug('just_installed_repository is installed: %s', pprint.pformat(attr.asdict(just_installed_repository)))

            just_installed_repositories.append(just_installed_repository)

    # log.debug('just_installed_repositories: %s', pprint.pformat(just_installed_repositories))

    return just_installed_repositories



import logging
import os
import shutil

import yaml

from ansible_galaxy import collection_info
from ansible_galaxy import collection_artifact_manifest
# from ansible_galaxy import collection_artifact_file_manifest
from ansible_galaxy import exceptions
from ansible_galaxy import install_info
from ansible_galaxy import requirements
from ansible_galaxy import role_metadata

from ansible_galaxy.models.repository_spec import RepositorySpec
from ansible_galaxy.models.repository import Repository

log = logging.getLogger(__name__)

# TODO: This loads a full repository object from disk. But that can
#       can be slow. Will likely need some sort of 'Summary' (to use cargo's term)
#       object that is just the important stuff (possibly cached somewhere?)
# aka, persistence of ansible_galaxy.models.repository


def load(data_or_file_object):
    repository_data = yaml.safe_load(data_or_file_object)
    return repository_data


def load_from_archive(repository_archive, namespace=None, installed=True):
    repo_tarfile = repository_archive.tar_file
    archive_path = repository_archive.info.archive_path

    # path_name = os.path.join(content_dir, namespace, name)
    path_name = repository_archive.info.top_dir

    manifest_filename = os.path.join(path_name, collection_artifact_manifest.COLLECTION_MANIFEST_FILENAME)
    manifest_data = None

    log.debug('Trying to extract %s from %s', manifest_filename, archive_path)

    try:
        mfd = repo_tarfile.extractfile(manifest_filename)
        if mfd:
            manifest_data = collection_artifact_manifest.load(mfd)
            log.debug('md: %s', manifest_data)
            log.debug('md.collection_info: %s', manifest_data.collection_info)
            log.debug('manifest_data.collection_info.name: %s', manifest_data.collection_info.name)
    except KeyError as e:
        log.warning('No %s found in archive: %s (Error: %s)', manifest_filename, archive_path, e)

    # load galaxy.yml
    galaxy_filename = os.path.join(path_name, collection_info.COLLECTION_INFO_FILENAME)

    collection_info_data = None

    try:
        gfd = repo_tarfile.extractfile(galaxy_filename)
        if gfd:
            collection_info_data = collection_info.load(gfd)
    except KeyError as e:
        log.warning('No %s found in archive: %s - %s', galaxy_filename, archive_path, e)
        # log.debug('No galaxy.yml collection info found for collection %s.%s: %s', namespace, name, e)

    # TODO/FIXME: what takes precedence?
    #           - the dir name in the archive that a collection lives in ~/.ansible/content/my_ns/my_name
    #           - Or the namespace/name from galaxy.yml?
    # log.debug('collection_info_data: %s', collection_info_data)

    col_info = None
    if manifest_data:
        col_info = manifest_data.collection_info
        log.debug('md.col_info: %s', col_info)
    elif collection_info_data:
        col_info = collection_info_data
    else:
        raise exceptions.GalaxyArchiveError('No galaxy collection info or manifest found in %s', archive_path)

    log.debug('col_info: %s', col_info)

    # FIXME: change collectionInfo to have separate name/namespace so we dont have to 'parse' the name
    # repo_spec = repository_spec.repository_spec_from_string(col_info.name, namespace_override=namespace)
    # spec_data = repository_spec_parse.parse_string(col_info.name)

    # log.debug('spec_data: %s', spec_data)
    # log.debug('repo_spec: %s', repo_spec)

    # Build a repository_spec of the repo now so we can pass it things like requirements.load()
    # that need to know what requires something
    # if we specify a namespace, use it otherwise use the info from galaxy.yml
    repo_spec = RepositorySpec(namespace=namespace or col_info.namespace,
                               name=col_info.name,
                               version=col_info.version,
                               spec_string=archive_path,
                               # fetch_method=None,
                               src=archive_path)

    log.debug('repo spec from %s: %r', archive_path, repo_spec)

    requirements_list = []
    requirements_list = requirements.from_requirement_spec_strings(col_info.dependencies,
                                                                   repository_spec=repo_spec)

    repository = Repository(repository_spec=repo_spec,
                            path=None,
                            installed=installed,
                            requirements=requirements_list,
                            # Assuming this is a collection artifact, FIXME if we support role artifacts
                            dependencies=[])

    log.debug('repository: %s', repository)

    return repository


def load_from_dir(content_dir, namespace, name, installed=True):
    # TODO: or artifact

    path_name = os.path.join(content_dir, namespace, name)
    # TODO: add trad role or collection detection rules here
    #       Or possibly earlier so we could call 'collection' loading
    #       code/class or trad-role-as-collection loading code/class
    #       and avoid intermingly the impls.
    #       Maybe:
    #       if more than one role in roles/ -> collection

    if not os.path.isdir(path_name):
        log.debug('The directory %s does not exist, unable to load a Repository from it', path_name)
        return None

    requirements_list = []

    # Now look for any install_info for the repository
    install_info_data = None
    install_info_filename = os.path.join(path_name, 'meta/.galaxy_install_info')
    try:
        with open(install_info_filename, 'r') as ifd:
            install_info_data = install_info.load(ifd)
    except EnvironmentError as e:
        log.warning('Unable to find or load meta/.galaxy_install_info for repository %s.%s: %s', namespace, name, e)

    # TODO: figure out what to do if the version from install_info conflicts with version
    #       from galaxy.yml etc.
    install_info_version = getattr(install_info_data, 'version', None)

    # Try to load a MANIFEST.json if we have one

    manifest_filename = os.path.join(path_name, collection_artifact_manifest.COLLECTION_MANIFEST_FILENAME)
    manifest_data = None

    try:
        with open(manifest_filename, 'r') as mfd:
            manifest_data = collection_artifact_manifest.load(mfd)
    except EnvironmentError:
        # log.debug('No galaxy.yml collection info found for collection %s.%s: %s', namespace, name, e)
        pass

    # # TODO/FIXME: do we even need to load file_manifest here?
    # file_manifest_filename = os.path.join(path_name, collection_artifact_file_manifest.COLLECTION_FILE_MANIFEST_FILENAME)
    # file_manifest_data = None

    # try:
    #     with open(file_manifest_filename, 'r') as mfd:
    #         file_manifest_data = collection_artifact_file_manifest.load(mfd)
    # except EnvironmentError:
    #     # log.debug('No galaxy.yml collection info found for collection %s.%s: %s', namespace, name, e)
    #     pass

    # load galaxy.yml
    galaxy_filename = os.path.join(path_name, collection_info.COLLECTION_INFO_FILENAME)

    collection_info_data = None

    try:
        with open(galaxy_filename, 'r') as gfd:
            if gfd:
                collection_info_data = collection_info.load(gfd)
    except EnvironmentError:
        # log.debug('No galaxy.yml collection info found for collection %s.%s: %s', namespace, name, e)
        pass

    # Now try the repository as a role-as-collection
    # FIXME: For a repository with one role that matches the collection name and doesn't
    #        have a galaxy.yml, that's indistinguishable from a role-as-collection
    # FIXME: But in theory, if there is more than one role in roles/, we should skip this

    role_dir_path = os.path.join(path_name, 'roles')
    role_name = '%s.%s' % (namespace, name)

    role_meta_main = role_metadata.load_from_dir(dirname=role_dir_path,
                                                 role_name=role_name)

    log.debug('role_meta_main: %s', role_meta_main)

    # Prefer version from install_info, but for a editable installed, there may be only galaxy version
    installed_version = install_info_version
    if manifest_data:
        installed_version = manifest_data.collection_info.version
    elif collection_info_data:
        installed_version = collection_info_data.version
    # if role_meta_main:
    #    installed_version = installed_version or role_meta_main.version

    # TODO/FIXME: what takes precedence?
    #           - the dir names a collection lives in ~/.ansible/content/my_ns/my_name
    #           - Or the namespace/name from galaxy.yml?
    # log.debug('collection_info_data: %s', collection_info_data)

    # Build a repository_spec of the repo now so we can pass it things like requirements.load()
    # that need to know what requires something
    repository_spec = RepositorySpec(namespace=namespace,
                                     name=name,
                                     version=installed_version)

    # The current galaxy.yml 'dependencies' are actually 'requirements' in ansible/ansible terminology
    # (ie, install-time)
    if collection_info_data:
        collection_requires = requirements.from_dependencies_dict(collection_info_data.dependencies,
                                                                  repository_spec=repository_spec)
        requirements_list.extend(collection_requires)

    # TODO: add reqs from MANIFEST.json
    # TODO: add requirements loaded from galaxy.yml
    # TODO: should the requirements in galaxy.yml be plain strings or dicts?
    # TODO: should there be requirements in galaxy.yml at all? in liue of requirements.yml
    # collection_info_requirements = []

    requirements_filename = os.path.join(path_name, 'requirements.yml')

    try:
        with open(requirements_filename, 'r') as rfd:
            requirements_list.extend(requirements.load(rfd, repository_spec=repository_spec))
    except EnvironmentError:
        # log.debug('No requirements.yml was loaded for repository %s.%s: %s', namespace, name, e)
        pass

    # TODO: if there are other places to load dependencies (ie, runtime deps) we will need
    #       to load them and combine them with role_depenency_specs

    # TODO/FIXME: load deps from MANIFEST.json if it exists (and prefer it over galaxy.yml)

    role_dependency_specs = []
    if role_meta_main:
        role_dependency_specs = role_meta_main.dependencies

    repository = Repository(repository_spec=repository_spec,
                            path=path_name,
                            installed=installed,
                            requirements=requirements_list,
                            dependencies=role_dependency_specs)

    log.debug('Repository %s loaded from %s', repository.repository_spec.label, path_name)

    return repository


def remove(installed_repository):
    log.info("Removing installed repository: %s", installed_repository)

    # editable installs are symlinks
    if os.path.islink(installed_repository.path):
        log.info('Removing the symlink at %s to %s',
                 installed_repository.path,
                 os.readlink(installed_repository.path))
        os.unlink(installed_repository.path)
        return True

    try:
        shutil.rmtree(installed_repository.path)
        return True
    except EnvironmentError as e:
        log.warning('Unable to rm the directory "%s" while removing installed repo "%s": %s',
                    installed_repository.path,
                    installed_repository.label,
                    e)
        log.exception(e)
        raise



import logging
import os
import tempfile
import pytest

from ansible_galaxy import repository_spec
from ansible_galaxy import exceptions
from ansible_galaxy.models.repository_spec import RepositorySpec, FetchMethods

log = logging.getLogger(__name__)

repo_spec_from_string_cases = \
    [
        # {'spec': 'geerlingguy.apache',
        # 'expected': RepositorySpec(name='apache', namespace='geerlingguy')},
        {'spec': 'geerlingguy.apache,2.1.1',
         'expected': RepositorySpec(name='apache', namespace='geerlingguy', version='2.1.1')},
        # {'spec': 'geerlingguy.apache',
        # 'expected': RepositorySpec(name='apache', namespace='geerlingguy', version=None)},
        # {'spec': 'testing.ansible-testing-content',
        # 'expected': RepositorySpec(name='ansible-testing-content', namespace='testing')},
        # {'spec': 'testing.ansible-testing-content,name=testing-content',
        # 'expected': RepositorySpec(name='testing-content', namespace='testing')},
        # {'spec': 'alikins.awx',
        # 'expected': RepositorySpec(name='awx', namespace='alikins')},
        {'spec': 'testing.ansible-testing-content,1.2.3,name=testing-content',
         'expected': RepositorySpec(name='testing-content', namespace='testing', version='1.2.3')},
        {'spec': 'testing.ansible-testing-content,1.2.3,also-testing-content,stuff',
         'expected': RepositorySpec(name='also-testing-content', namespace='testing', version='1.2.3')},
        # for git/tar/url, we dont try to guess the namespace, so the expected result is namespace=None
        # here. cli adds a namespace from --namespace here.
        {'spec': 'git+https://github.com/geerlingguy/ansible-role-apache.git,version=2.0.0',
         'expected': RepositorySpec(name='ansible-role-apache', namespace=None, version='2.0.0',
                                    scm='git')},
        {'spec': 'git+https://mazertestuser@github.com/geerlingguy/ansible-role-apache.git,version=2.0.0',
         'expected': RepositorySpec(name='ansible-role-apache', namespace=None, version='2.0.0',
                                    scm='git', fetch_method=FetchMethods.SCM_URL)},
        # A path to a file without a dot in it's name. It's path will include where the tests are run from
        # so specify a ',name=' to provide a predictable name (otherwise it would be the full path)
        # TODO: local file will attempt to load the contents now, so this test case doesn't make sense without mocking
        #       out something to exist at the path
        # {'spec': '%s,name=the_license' % os.path.normpath(os.path.join(os.path.dirname(__file__), '../../LICENSE')),
        # 'expected': RepositorySpec(name='the_license', namespace=None,
        #                            fetch_method=FetchMethods.LOCAL_FILE)},
        # {'spec': 'https://docs.ansible.com,name=the_docs',
        # 'expected': RepositorySpec(name='the_docs', namespace=None,
        #                            scm=None, fetch_method=FetchMethods.REMOTE_URL)},
        # 'foo',
        # 'foo,1.2.3',
        # 'foo,version=1.2.3',
        # 'foo,1.2.3,somename',
        # 'foo,1.2.3,name=somename',
        # 'foo,1.2.3,somename,somescm',
        # 'foo,1.2.3,somename,somescm,someextra'
    ]


@pytest.fixture(scope='module',
                params=repo_spec_from_string_cases,
                ids=[x['spec'] for x in repo_spec_from_string_cases])
def repository_spec_case(request):
    yield request.param


def test_repository_spec_from_string(repository_spec_case):
    result = repository_spec.repository_spec_from_string(repository_spec_case['spec'])
    log.debug('spec=%s result=%s exp=%s', repository_spec_case['spec'], result, repository_spec_case['expected'])

    assert result == repository_spec_case['expected']


def test_repository_spec_editable():
    tmpdir = tempfile.mkdtemp()
    expected = os.path.basename(tmpdir)
    log.debug('expected: %s', expected)

    result = repository_spec.repository_spec_from_string(tmpdir, namespace_override='my_namespace', editable=True)
    os.rmdir(tmpdir)

    log.debug('result: %r', result)

    assert result.name == expected
    assert result.namespace == 'my_namespace'
    assert result.version is None
    assert result.src == tmpdir
    assert result.fetch_method == 'EDITABLE'


@pytest.mark.xfail(raises=exceptions.GalaxyError)
def test_repository_spec_fail():
    repository_spec.repository_spec_from_string('foo.')


@pytest.mark.xfail(raises=exceptions.GalaxyError)
def test_repository_editable_fail():
    repository_spec.repository_spec_from_string('foo', editable=True)





import logging

import attr

from ansible_galaxy.models.collection_info import CollectionInfo
# from ansible_galaxy.models.collection_artifact_file_manifest_info import \
#    CollectionArtifactFileManifestInfo
from ansible_galaxy.models.collection_artifact_file import CollectionArtifactFile

log = logging.getLogger(__name__)


# see https://github.com/ansible/galaxy/issues/957
@attr.s(frozen=True)
class CollectionArtifactManifest(object):
    collection_info = attr.ib(type=CollectionInfo,
                              validator=attr.validators.instance_of(CollectionInfo))

    file_manifest_file = attr.ib(default=None,
                                 type=CollectionArtifactFile,
                                 validator=attr.validators.optional(attr.validators.instance_of(CollectionArtifactFile)))

    format = attr.ib(default=1)

    # TODO: add a attr with info about the FILES.* file,
    #       files_info_path, files_info_chksum
    #       We need a chksum here, so a signed/verified MANIFEST.json
    #       can be used to verified chksum of FILES.JSON and it can
    #       verify chksums of the rest of the bundled files

    # maybe 'contents' in the future
    #
    # maybe build_info = attr.ib(type=CollectionArtifactBuildInfo)
    #   CollectionArtifactBuildInfo has 'build_date', 'build_tool'
    # created_with =
    # when/build_date =
    # files = []  list of CollectionArtifactManifestFileInfo  (eek... long name)



import logging
import pytest

from ansible_galaxy import matchers
from ansible_galaxy.models.repository import Repository
from ansible_galaxy.models.repository_spec import RepositorySpec

log = logging.getLogger(__name__)


def CR(namespace=None, name=None):
    cs = RepositorySpec(namespace=namespace,
                        name=name,
                        version='1.2.3')
    return Repository(repository_spec=cs)


@pytest.mark.parametrize("matcher_class,matcher_args,candidates, expected", [
    (matchers.MatchAll, tuple(), [1], True),
    (matchers.MatchAll, tuple(), ['a', 'b', 3],  True),
    (matchers.MatchNone, tuple(), ['a', 'b', 3],  False),
    (matchers.MatchNames, (['name1', 'name2'], ), [CR(name='name1')], True),
    (matchers.MatchNames, (['name1', 'name2'], ), [CR(name='blargh')], False),
    (matchers.MatchLabels,
     (['ns1.name1'], ),
     [CR(namespace='ns1', name='name1')],
     True),
    (matchers.MatchLabels,
     (['ns1.name1'], ),
     [CR(namespace='ns1', name='name2')],
     False),
    (matchers.MatchLabels,
     (['ns1.name1'], ),
     [CR(namespace='ns2', name='name1')],
     False),
    (matchers.MatchLabels,
     (['ns1.name1'], ),
     [CR(namespace='ns2', name='name2')],
     False),
    (matchers.MatchNamespacesOrLabels,
     (['ns1', 'offlabel'],),
     [CR(namespace='ns1')],
     True),
    (matchers.MatchNamespacesOrLabels,
     (['ns1', 'offlabel'],),
     [CR(namespace='ns1', name='name1')],
     True),
    (matchers.MatchNamespacesOrLabels,
     (['ns1.name1'],),
     [CR(namespace='ns1', name='name1')],
     True),
    (matchers.MatchNamespacesOrLabels,
     (['ns1.name1'],),
     [CR(namespace='ns1', name='name2')],
     False),
    (matchers.MatchNamespacesOrLabels,
     (['ns1.name1'],),
     [CR(namespace='ns2', name='name1')],
     False),
    (matchers.MatchNamespacesOrLabels,
     (['ns2.name2'],),
     [CR(namespace='ns1', name='name1')],
     False)

])
def test_match_all(matcher_class, matcher_args, candidates, expected):
    log.debug('matcher_class=%s matcher_args=%s candidate=%s, expected=%s', matcher_class, matcher_args, candidates, expected)

    matcher = matcher_class(*matcher_args)
    for candidate in candidates:
        res = matcher(candidate)

        log.debug('res: %s, matcher=%s, candidate=%s, expected=%s', res, matcher, candidate, expected)
        assert res is expected, 'res %s for matcher=%s candidate=%s was not %s' % (res, matcher, candidate, expected)




import datetime
import logging
import os

import semantic_version

from ansible_galaxy import repository_archive
from ansible_galaxy.models.repository_spec import RepositorySpec
from ansible_galaxy.models.install_destination import InstallDestinationInfo
from ansible_galaxy.models.install_info import InstallInfo
from ansible_galaxy.models.installation_results import InstallationResults
from ansible_galaxy.models.repository_archive import RepositoryArchiveInfo
from ansible_galaxy.models.repository_archive import CollectionRepositoryArtifactArchive
from ansible_galaxy.actions import build
from ansible_galaxy.models.build_context import BuildContext

log = logging.getLogger(__name__)


def display_callback(msg, **kwargs):
    log.debug(msg)


def build_repo_artifact(galaxy_context_, tmp_dir):
    output_path = tmp_dir.mkdir('mazer_test_repository_archive_test_build')

    collection_path = os.path.join(os.path.dirname(__file__), 'collection_examples/hello')
    build_context = BuildContext(collection_path, output_path=output_path.strpath)
    ret = build._build(galaxy_context_, build_context, display_callback)

    log.debug('ret: %s', ret)

    return ret


def test_install(galaxy_context, tmpdir):
    built_res = build_repo_artifact(galaxy_context, tmpdir)
    archive_path = built_res['build_results'].artifact_file_path

    repo_archive = repository_archive.load_archive(archive_path)

    log.debug('repo_archive: %s', repo_archive)

    repo_spec = RepositorySpec(namespace='some_namespace',
                               name='some_name',
                               version='1.2.3')

    namespaced_repository_path = '%s/%s' % (repo_spec.namespace,
                                            repo_spec.name)

    extract_archive_to_dir = os.path.join(galaxy_context.content_path,
                                          namespaced_repository_path,
                                          '')

    install_info_path = os.path.join(galaxy_context.content_path,
                                     namespaced_repository_path,
                                     'meta/.galaxy_install_info')

    destination_info = InstallDestinationInfo(destination_root_dir=galaxy_context.content_path,
                                              repository_spec=repo_spec,
                                              extract_archive_to_dir=extract_archive_to_dir,
                                              namespaced_repository_path=namespaced_repository_path,
                                              install_info_path=install_info_path,
                                              force_overwrite=True,
                                              editable=False)

    res = repository_archive.install(repo_archive, repo_spec, destination_info, display_callback=display_callback)

    log.debug('res: %s', res)

    assert isinstance(res, InstallationResults)
    assert isinstance(res.install_info, InstallInfo)
    assert isinstance(res.install_info.version, semantic_version.Version)
    assert isinstance(res.installed_datetime, datetime.datetime)


def test_load_from_archive(galaxy_context, tmpdir):
    built_res = build_repo_artifact(galaxy_context, tmpdir)
    archive_path = built_res['build_results'].artifact_file_path

    res = repository_archive.load_archive(archive_path)

    log.debug('res: %s', res)

    assert isinstance(res, CollectionRepositoryArtifactArchive)
    assert isinstance(res.info, RepositoryArchiveInfo)

    # CollectionRepositoryArtifactArchive(info=RepositoryArchiveInfo(archive_type='multi-content-artifact', top_dir='greetings_namespace.hello-11.11.11'
    assert res.info.archive_type == 'multi-content-artifact'

    assert res.info.top_dir == ''




import logging
import os
import subprocess

import pytest

log = logging.getLogger(__name__)


def popen_git_command(git_command, git_repo_path, expected_return_code=0):
    log.debug('Running command: %s in %s', git_command, git_repo_path)

    try:
        popen = subprocess.Popen(git_command, cwd=git_repo_path.strpath, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    except Exception as exc:
        log.debug(exc, exc_info=True)
        raise

    stdout, stderr = popen.communicate()
    output = "stdout:\n%s\n\nstderr:\n%s" % (stdout, stderr)

    rc = popen.returncode
    if rc != expected_return_code:
        raise Exception("git command '%s' failed in directory %s (rc=%s, expected_rc=%s)\n%s" %
                        (' '.join(git_command), git_repo_path.strpath, rc, expected_return_code, output))


@pytest.fixture(scope="session")
def tmp_git_repo(tmpdir_factory):
    git_repo_path = tmpdir_factory.mktemp('mazer_unit_test_git_repo')

    git_init_command = ['git', 'init', '.']
    log.debug('Creating tmp git repo in %s', git_repo_path.strpath)

    popen_git_command(git_init_command, git_repo_path)

    git_repo_path.mkdir('some_dir')

    some_file_bases = ['some_dir/another_file', 'some_file', 'README.md', 'build_artifact.pyc']
    some_files = [os.path.join(git_repo_path.strpath, file_base) for file_base in some_file_bases]

    for some_file in some_files:
        with open(some_file, 'w') as somefd:
            somefd.write('whatever\n')

    git_add_command = ['git', 'add', '.']
    popen_git_command(git_add_command, git_repo_path)

    # have to setup git username/email before we can commit. Normally git will try to guess
    # here, based on $HOME, but Since tox envs do not have a $HOME set (or a home dir at all) we get
    # git errors here because of no configured username/name.

    git_email_config_command = ['git', 'config', 'user.email', "mazer_test_user@example.com"]
    popen_git_command(git_email_config_command, git_repo_path)

    git_username_config_command = ['git', 'config', 'user.name', 'Mazer "Corn Dog" Rackham']
    popen_git_command(git_username_config_command, git_repo_path)

    # TODO: setup gitconfig for user.email etc since there is no HOME set in tox envs
    git_commit_command = ['git', 'commit', '--message', 'test commit wip']
    popen_git_command(git_commit_command, git_repo_path)

    return git_repo_path



import logging

import attr


log = logging.getLogger(__name__)


@attr.s(frozen=True)
class RoleMetadata(object):
    '''The info that is found in a role meta/main.yml file'''
    name = attr.ib(default=None)

    author = attr.ib(default=None)
    description = attr.ib(default=None)
    company = attr.ib(default=None)
    license = attr.ib(default=None)

    # behaviorial
    min_ansible_version = attr.ib(default=None, converter=str)
    min_ansible_container_version = attr.ib(default=None, converter=str)
    allow_duplicates = attr.ib(default=False)

    issue_tracker = attr.ib(default=None)
    github_branch = attr.ib(default=None)

    # TODO: validate list items are text
    galaxy_tags = attr.ib(factory=list)

    # TODO: a Platform model if needed
    platforms = attr.ib(factory=list)

    cloud_platforms = attr.ib(factory=list)

    # TODO: a role/content Dependency model
    dependencies = attr.ib(factory=list)



import logging
import pprint

import attr
from attr.exceptions import NotAnAttrsClassError

import six

log = logging.getLogger(__name__)


def is_attr(obj):
    if isinstance(obj, six.class_types):
        try:
            attr.fields(obj)
        except NotAnAttrsClassError:
            return False
        return False

    if getattr(obj.__class__, "__attrs_attrs__", None):
        return True

    if getattr(obj, "__attrs_attrs__", None):
        return True

    return False


def pf_attr(obj):
    if not is_attr(obj):
        return pprint.pformat(obj)

    return pprint.pformat(attr.fields_dict(obj))



import logging
import os

log = logging.getLogger(__name__)


class BaseFetch(object):
    fetch_method = 'base'

    def __init__(self):
        self.local_path = None
        self.cleanup_tmp_files = True

        # remote_resource is whatever was fetched, ie a url or galaxy content spec
        # just an identifier to use in messages
        self.remote_resource = None

    def find(self):
        '''Lookup the requested spec, and if found, return a RepositorySpec

        More or less resolve the ambiquious RequirementSpec to a particular concrete
        RepositorySpec'''
        raise NotImplementedError

    def fetch(self, find_results=None):
        '''Get the content archive, save it to a file locally and return the path to that file'''
        raise NotImplementedError

    def cleanup(self):
        if not self.cleanup_tmp_files:
            log.info("cleanup_tmp_files is false, Not removing the tmp file %s fetched from %s",
                     self.local_path, self.remote_resource)
            return

        log.debug("Removing the tmp file %s fetched from %s",
                  self.local_path, self.remote_resource)

        try:
            os.unlink(self.local_path)
        except (OSError, IOError) as e:
            log.warning('Unable to remove tmp file (%s): %s' % (self.local_path, str(e)))



import logging
import os
import pprint
import tarfile

from ansible_galaxy import build
from ansible_galaxy.models.build_context import BuildContext
from ansible_galaxy.models.collection_info import CollectionInfo
from ansible_galaxy.models.collection_artifact_manifest import CollectionArtifactManifest


log = logging.getLogger(__name__)
pf = pprint.pformat


def display_callback(*args, **kwargs):
    print(args)


def _build_context(collection_path=None, output_path=None):
    collection_path = collection_path or os.path.join(os.path.dirname(__file__), 'collection_examples/hello')
    log.debug('collection_path: %s', collection_path)
    log.debug('output_path: %s', output_path)

    return BuildContext(collection_path=collection_path,
                        output_path=output_path)


def _collection_info(namespace=None, name=None, version=None, authors=None):
    # name = name or 'some_namespace.some_name'
    namespace = namespace or 'some_namespace'
    name = name or 'some_name'
    version = version or '1.2.3'
    authors = authors or ['Rex Chapman']
    description = "Unit testing thing"
    test_license = 'GPL-3.0-or-later'

    return CollectionInfo(namespace=namespace, name=name, version=version, authors=authors, description=description,
                          license=test_license)


def test_build(tmpdir):
    temp_dir = tmpdir.mkdir('mazer_test_build_unit_test')
    build_context = _build_context(output_path=temp_dir.strpath)
    collection_info = _collection_info()
    build_ = build.Build(build_context, collection_info)

    log.debug('build_: %s', build_)
    assert isinstance(build_, build.Build)
    assert isinstance(build_.build_context, BuildContext)


def test_build_run(tmpdir):
    temp_dir = tmpdir.mkdir('mazer_test_build_run_unit_test')
    build_context = _build_context(output_path=temp_dir.strpath)
    collection_info = _collection_info()
    build_ = build.Build(build_context, collection_info)

    res = build_.run(display_callback)

    assert isinstance(res, build.BuildResult)
    assert isinstance(res.manifest, CollectionArtifactManifest)
    assert isinstance(res.manifest.collection_info, CollectionInfo)
    assert isinstance(res.file_manifest.files, list)

    assert res.manifest.collection_info == collection_info

    for manifest_file in res.file_manifest.files:
        if manifest_file.ftype == 'file':
            assert manifest_file.chksum_type == 'sha256'

    file_names = [x.name for x in res.file_manifest.files if x.ftype == 'file']
    dir_names = [x.name for x in res.file_manifest.files if x.ftype == 'dir']

    log.debug('file_names: %s', file_names)
    log.debug('dir_names: %s', dir_names)

    assert 'roles/some_role/meta/main.yml' in file_names
    assert 'galaxy.yml' in file_names
    assert 'roles' in dir_names
    assert 'roles/some_role/defaults' in dir_names

    assert isinstance(res.messages, list)

    # verify the output file got created
    artifact_file_path = res.artifact_file_path
    assert os.path.isfile(artifact_file_path)
    assert tarfile.is_tarfile(artifact_file_path)

    log.debug('results.errors: %s', res.errors)



import logging
import os

from ansible_galaxy import exceptions
from ansible_galaxy import galaxy_repository_spec
from ansible_galaxy.models.repository_spec import FetchMethods

log = logging.getLogger(__name__)


def repo_url_to_repo_name(repo_url):
    # gets the role name out of a repo like
    # http://git.example.com/repos/repo.git" => "repo"

    if '://' not in repo_url and '@' not in repo_url:
        return repo_url
    trailing_path = repo_url.split('/')[-1]
    if trailing_path.endswith('.git'):
        trailing_path = trailing_path[:-4]
    if trailing_path.endswith('.tar.gz'):
        trailing_path = trailing_path[:-7]
    if ',' in trailing_path:
        trailing_path = trailing_path.split(',')[0]
    return trailing_path


def split_kwarg(spec_string, valid_keywords):
    if '=' not in spec_string:
        return (None, spec_string)

    parts = spec_string.split('=', 1)

    if parts[0] in valid_keywords:
        return (parts[0], parts[1])

    raise exceptions.GalaxyClientError('The repository spec uses an unsuppoted keyword: %s' % spec_string)


def split_comma(spec_string, valid_keywords):
    # res = []
    comma_parts = spec_string.split(',')
    for comma_part in comma_parts:
        kw_parts = split_kwarg(comma_part, valid_keywords)
#        log.debug('kw_parts: %s', kw_parts)
        yield kw_parts


def split_repository_spec(spec_string, valid_keywords):
    comma_splitter = split_comma(spec_string, valid_keywords)
    info = {}
    for kw in valid_keywords:
        try:
            key, value = next(comma_splitter)
        except StopIteration:
            return info

        if key:
            info[key] = value
        else:
            info[kw] = value

    return info


def parse_string(repository_spec_text, valid_keywords=None):
    '''Given a text/str object describing a galaxy repository, parse it.

    And return a dict with keys: 'name', 'src', 'scm', 'version'
    '''

    valid_keywords = valid_keywords or ('src', 'version', 'name', 'namespace', 'scm')
    data = {'src': None,
            'name': None,
            'namespace': None,
            # 'sub_name': None,
            'version': None,
            'scm': None,
            'spec_string': None}

    data['spec_string'] = repository_spec_text

    split_data = split_repository_spec(repository_spec_text, valid_keywords)

    # src = split_data.pop('src')
    data.update(split_data)
    # data['src'] = src

    return data


def is_scm(repository_spec_string):
    if '://' in repository_spec_string or '@' in repository_spec_string:
        return True

    return False


def chose_repository_fetch_method(repository_spec_string):
    # log.debug('repository_spec_string: %s', repository_spec_string)

    if is_scm(repository_spec_string):
        # create tar file from scm url
        return FetchMethods.SCM_URL

    comma_parts = repository_spec_string.split(',', 1)
    potential_filename = comma_parts[0]
    if os.path.isfile(potential_filename):
        # installing a local tar.gz
        return FetchMethods.LOCAL_FILE

    if '://' in repository_spec_string:
        return FetchMethods.REMOTE_URL

    # if it doesnt look like anything else, assume it's galaxy
    return FetchMethods.GALAXY_URL


def resolve(data):
    src = data['src']
    if data['name'] is None:
        scm_name = repo_url_to_repo_name(src)
        data['name'] = scm_name
        if '+' in src:
            (scm_url, scm_src) = src.split('+', 1)
            data['scm'] = scm_url
            data['src'] = scm_src

    # split the name on '.' and recombine the first 1 or 2
    name_parts = data['name'].split('.')
    new_name_parts = []
    new_name_parts.append(name_parts.pop(0))

    # we may not have a second part to the name
    try:
        new_name_parts.append(name_parts.pop(0))
    except IndexError:
        pass

    # combine the name parts, which may be one or two parts
    data['name'] = '.'.join(new_name_parts)

    return data


def spec_data_from_string(repository_spec_string, resolver=None):
    fetch_method = chose_repository_fetch_method(repository_spec_string)

    # log.debug('fetch_method: %s', fetch_method)

    spec_data = parse_string(repository_spec_string)
    spec_data['fetch_method'] = fetch_method

    # log.debug('spec_data: %s', spec_data)

    # use passed in resolver if provided, otherwise assume 'resolve' is correct
    # but override if it looks like a galaxy requests
    if resolver is None:
        resolver = resolve
        if fetch_method == FetchMethods.GALAXY_URL:
            resolver = galaxy_repository_spec.resolve

    # log.debug('resolver: %s', resolver)

    resolved_name = resolver(spec_data)

    # log.debug('resolved_name: %s', resolved_name)

    spec_data.update(resolved_name)

    return spec_data



import logging
import os
import sys

from ansible_galaxy.utils.text import to_text

log = logging.getLogger(__name__)

VERSION_FIELDS = ('name', 'version', 'config_file', 'uname', 'executable_location', 'python_version', 'python_executable')


def version_data(config_file_path, cli_version, argv):
    data = {}

    data['name'] = 'mazer'
    data['version'] = cli_version
    data['executable_location'] = argv[0]
    data['uname'] = u', '.join(os.uname())

    sys_ver = u"%s" % ''.join(sys.version.splitlines())

    data['python_version'] = sys_ver
    data['python_executable'] = sys.executable

    if config_file_path:
        data['config_file'] = to_text(config_file_path)
    else:
        data['config_file'] = u'No config file found; using defaults'

    return data


def version_repr(version_data):
    lines = []

    for field in VERSION_FIELDS:
        lines.append(u'%s = %s' % (field, version_data.get(field, '')))

    buf = u'\n'.join(lines)
    return buf


def version(config_file_path, cli_version, display_callback=None):
    version_buf = version_repr(version_data(config_file_path, cli_version, sys.argv))
    display_callback(version_buf)
    return 0



import logging

from ansible_galaxy.models import role_metadata

import attr

log = logging.getLogger(__name__)


def test_init_empty():
    role_md = role_metadata.RoleMetadata()

    log.debug('role_md: %s', role_md)

    assert isinstance(role_md, role_metadata.RoleMetadata)


def test_basic():
    role_md = role_metadata.RoleMetadata(name='some_role',
                                         author='alikins@redhat.com',
                                         description='some role that does stuff',
                                         company='Red Hat',
                                         license='GPLv3',
                                         galaxy_tags=['stuff', 'nginx', 'system', 'devel'])

    log.debug('role_md: %s', role_md)
    assert isinstance(role_md, role_metadata.RoleMetadata)

    assert role_md.name == 'some_role'
    assert role_md.author == 'alikins@redhat.com'
    assert 'stuff' in role_md.galaxy_tags
    assert isinstance(role_md.galaxy_tags, list)
    assert role_md.allow_duplicates is False


def test_equal():
    role_md1 = role_metadata.RoleMetadata(name='some_role')
    role_md1a = role_metadata.RoleMetadata(name='some_role')
    role_md2 = role_metadata.RoleMetadata(name='a_different_role')

    assert role_md1 == role_md1a
    assert role_md1a == role_md1

    assert not role_md1 == role_md2
    assert not role_md2 == role_md1

    assert role_md1 != role_md2
    assert role_md2 != role_md1


def test_asdict():
    role_md = role_metadata.RoleMetadata(name='some_role',
                                         author='alikins@redhat.com',
                                         description='some role that does stuff',
                                         company='Red Hat',
                                         license='GPLv3',
                                         galaxy_tags=['stuff', 'nginx', 'system', 'devel'])

    log.debug('role_md: %s', role_md)
    role_dict = attr.asdict(role_md)

    assert isinstance(role_dict, dict)
    assert role_dict['name'] == role_md.name == 'some_role'
    assert role_dict['galaxy_tags'] == role_md.galaxy_tags == ['stuff', 'nginx', 'system', 'devel']



import os
import tarfile

import attr


@attr.s(frozen=True)
class RepositoryArchiveInfo(object):
    requires_meta_main = False

    archive_type = attr.ib()
    top_dir = attr.ib()
    archive_path = attr.ib(default=None)

    # download url?
    # file path?
    # checksum?
    # signature?


@attr.s(frozen=True)
class RepositoryArchive(object):
    info = attr.ib(type=RepositoryArchiveInfo)
    tar_file = attr.ib(type=tarfile.TarFile, default=None)

    META_INSTALL = os.path.join('meta', '.galaxy_install_info')

    def repository_dest_root_subpath(self, repository_name):
        '''The relative path inside the installed content where extract should consider the root

        A collection archive for 'my_namespace.my_content' will typically be extracted to
        '~/.ansible/content/my_namespace/my_content' in which case the repository_dest_root_subpath
        should return just '/'.

        But Role archives will be extracted into a 'roles' sub dir of the typical path.
        ie, a 'my_namespace.my_role' role archive will need to be extracted to
        '~/.ansible/content/my_namespace/roles/my_role/' in which case the repository_dest_root_subpatch
        should return 'roles/my_roles' (ie, 'roles/%s' % content_name)
        '''
        return ''


@attr.s(frozen=True)
class TraditionalRoleRepositoryArchive(RepositoryArchive):
    ROLES_SUBPATH = 'roles'

    def repository_dest_root_subpath(self, repository_name):
        '''Traditional role archive repository gets installed into subpath of 'roles/CONTENT_NAME/'''
        return os.path.join(self.ROLES_SUBPATH, repository_name)


@attr.s(frozen=True)
class CollectionRepositoryArchive(RepositoryArchive):
    pass


@attr.s(frozen=True)
class CollectionRepositoryArtifactArchive(RepositoryArchive):
    pass



import logging

from ansible_galaxy.models import repository_spec
from ansible_galaxy.models import requirement
from ansible_galaxy.models import requirement_spec

log = logging.getLogger(__name__)


def test_requirement_op():
    ops = requirement.RequirementOps()
    assert isinstance(ops, requirement.RequirementOps)
    assert ops.EQ == '='


def test_requirement():
    repo_spec1 = repository_spec.RepositorySpec(namespace='mynamespace',
                                                name='myname',
                                                version='3.4.5')

    provides_spec2 = requirement_spec.RequirementSpec(namespace='othernamespace',
                                                      name='provides_something',
                                                      version_spec='==1.0.0')

    req = requirement.Requirement(repository_spec=repo_spec1, requirement_spec=provides_spec2,
                                  op=requirement.RequirementOps.EQ)

    assert isinstance(req, requirement.Requirement)
    assert req.repository_spec == repo_spec1
    assert req.requirement_spec == provides_spec2


def test_repository_spec_requirement_spec_cmp():
    repo_spec1 = repository_spec.RepositorySpec(namespace='ns',
                                                name='n',
                                                version='3.4.5')

    req_spec = requirement_spec.RequirementSpec(namespace='ns',
                                                name='n',
                                                version_spec='==3.4.5')

    req = requirement.Requirement(repository_spec=repo_spec1, requirement_spec=req_spec,
                                  op=requirement.RequirementOps.EQ)

    log.debug('req: %s', req)
    log.debug('repo_spec1: %s', repo_spec1)
    log.debug('req_spec: %s', req_spec)

    assert req.repository_spec == repo_spec1
    assert req.requirement_spec == req_spec



import logging

from ansible_galaxy import repository
from ansible_galaxy import installed_repository_db
from ansible_galaxy import matchers
from ansible_galaxy_cli import exceptions as cli_exceptions

log = logging.getLogger(__name__)


def remove_repository(installed_repository,
                      display_callback=None):
    log.debug('looking for repository %s to remove', installed_repository)

    log.debug('repository to remove: %s %s', installed_repository, type(installed_repository))

    try:
        res = repository.remove(installed_repository)
        if res:
            display_callback('- successfully removed %s' % installed_repository.label)
        else:
            display_callback('- %s is not installed, skipping.' % installed_repository.label)
    except Exception as e:
        log.exception(e)
        raise cli_exceptions.GalaxyCliError("Failed to remove installed repository %s: %s" %
                                            (installed_repository.label, str(e)))

    # FIXME: return code?  was always returning 0


def remove(galaxy_context,
           repository_spec_match_filter=None,
           display_callback=None):

    repository_spec_match_filter = repository_spec_match_filter or matchers.MatchNone()

    irdb = installed_repository_db.InstalledRepositoryDatabase(galaxy_context)

    for matched_repository in irdb.select(repository_spec_match_filter=repository_spec_match_filter):
        log.debug('removing %s', matched_repository)
        # content_info['content_data'].remove()
        remove_repository(matched_repository,
                          display_callback=display_callback)

    return 0



import collections
import logging

from ansible_galaxy.config import defaults
from ansible_galaxy.config import config_file

log = logging.getLogger(__name__)


class Config(object):
    def __init__(self):
        self.server = {}
        self.content_path = None
        self.global_content_path = None
        self.options = {}

    def as_dict(self):
        return collections.OrderedDict([
            ('server', self.server),
            ('content_path', self.content_path),
            ('global_content_path', self.global_content_path),
            ('options', self.options),
        ])

    @classmethod
    def from_dict(cls, data):
        inst = cls()
        inst.server = data.get('server', inst.server)
        inst.content_path = data.get('content_path', inst.content_path)
        inst.global_content_path = data.get('global_content_path', inst.global_content_path)
        inst.options = data.get('options', inst.options)
        return inst


def load(full_file_path):
    '''Load the yaml config file at full_file_path and create and return an instance of Config'''
    config_file_data = config_file.load(full_file_path)

    _default_conf_data = collections.OrderedDict(defaults.DEFAULTS)

    config_data = config_file_data or _default_conf_data

    for key in _default_conf_data:
        # If an expected key is not defined, set it to the default value
        if key not in config_data:
            config_data[key] = _default_conf_data[key]

    log.debug('config_data: %s', config_data)

    return Config.from_dict(config_data)


def save(config_obj, full_file_path):
    '''Save an instance of Config (config_obj) to full_file_path'''

    config_data = config_obj.as_dict()

    return config_file.save(config_data, full_file_path)



import logging

import attr
import semantic_version

from ansible_galaxy.utils.version import convert_string_to_semver
log = logging.getLogger(__name__)


# FIXME: do we have an enum like class for py2.6? worth a dep?
class FetchMethods(object):
    SCM_URL = 'SCM_URL'
    LOCAL_FILE = 'LOCAL_FILE'
    REMOTE_URL = 'REMOTE_URL'
    GALAXY_URL = 'GALAXY_URL'
    EDITABLE = 'EDITABLE'


@attr.s(frozen=True)
class RepositorySpec(object):
    '''The info used to identify and reference a galaxy content.

    For ex, 'testing.ansible-testing-content' will result in
    a RepositorySpec(name=ansible-testing-content, repo=ansible-testing-content,
                  namespace=testing, raw=testing.ansible-testing-content)'''
    namespace = attr.ib()
    name = attr.ib()
    version = attr.ib(type=semantic_version.Version,
                      # default=None,
                      converter=convert_string_to_semver,
                      validator=attr.validators.optional(attr.validators.instance_of(semantic_version.Version)))

    # only namespace/name/version are used for eq checks currently

    # TODO: add explicit attribute for repository 'source' (universe? provider?)
    #       ie, 'galaxy' or 'internal-dev' or 'local' etc. Maybe the combo of
    #       'fetch_method' and 'src' are close enough. Goal being to allow a
    #       repository to spec a requirement on another repository (namespace, name, version)
    #       as well as where it came from (galaxy, some github url, a specific server, etc)

    fetch_method = attr.ib(default=None, cmp=False)
    scm = attr.ib(default=None, cmp=False)
    spec_string = attr.ib(default=None, cmp=False)
    src = attr.ib(default=None, cmp=False)

    @property
    def label(self):
        return '%s.%s' % (self.namespace, self.name)

    @classmethod
    def from_dict(cls, data):
        instance = cls(namespace=data['namespace'],
                       name=data['name'],
                       version=data.get('version', None),
                       fetch_method=data.get('fetch_method', None),
                       scm=data.get('scm', None),
                       spec_string=data.get('spec_string', None),
                       src=data.get('src', None),
                       )
        return instance

    def __str__(self):
        return '{label}-{version}'.format(label=self.label, version=self.version)



import logging

log = logging.getLogger(__name__)


VALID_ROLE_SPEC_KEYS = [
    'src',
    'version',
    'role',
    'name',
    'scm',
]

# Galaxy Content Constants
CONTENT_PLUGIN_TYPES = (
    'module',
    'module_util',
    'plugin'
)
CONTENT_TYPES = CONTENT_PLUGIN_TYPES + ('role',)

CONTENT_TYPE_DIR_MAP = dict([(k, '%ss' % k) for k in CONTENT_TYPES])
TYPE_DIR_CONTENT_TYPE_MAP = dict([('%ss' % k, k) for k in CONTENT_TYPES])



import pytest


@pytest.fixture
def galaxy_context(tmpdir):
    # tmp_content_path = tempfile.mkdtemp()
    # FIXME: mock
    server = {'url': 'http://localhost:8000',
              'ignore_certs': False}
    content_dir = tmpdir.mkdir('collections')
    collections_dir = content_dir.mkdir('ansible_collections')
    from ansible_galaxy.models.context import GalaxyContext
    return GalaxyContext(server=server, content_path=collections_dir.strpath)



import logging
import os
import tempfile


from ansible_galaxy.fetch import local_file
from ansible_galaxy.models.repository_spec import FetchMethods
from ansible_galaxy.models.requirement_spec import RequirementSpec

log = logging.getLogger(__name__)


JSON_DATA = b'''
{
    "collection_info": {
        "namespace": "bcs",
        "name": "bcsping",
        "version": "0.0.1",
        "license": "GPL-2.0-or-later",
        "description": "i blatant copy and rename of the ping module",
        "repository": null,
        "documentation": null,
        "homepage": null,
        "issues": null,
        "authors": [
            "me"
        ],
        "tags": [],
        "readme": "README.md",
        "dependencies": []
    },
    "format": 1,
    "files": [
        {
            "name": ".",
            "ftype": "dir",
            "chksum_type": null,
            "chksum_sha256": null,
            "_format": 1
        },
        {
            "name": "galaxy.yml",
            "ftype": "file",
            "chksum_type": "sha256",
            "chksum_sha256": "728ec435b1717502c305593c7d07116eaacb17ebf714a316f7a925e222f9bb24",
            "_format": 1
        }
    ]
}
'''


def test_local_file_fetch(mocker):
    tmp_file_fo = tempfile.NamedTemporaryFile(prefix='tmp', suffix='.tar.gz', delete=True)
    log.debug('tmp_file_fo.name=%s tmp_file=%s', tmp_file_fo.name, tmp_file_fo)

    requirement_spec_ = RequirementSpec(namespace='namespace', name='name', version_spec='==1.2.3',
                                        fetch_method=FetchMethods.LOCAL_FILE,
                                        src=tmp_file_fo.name)

    mocker.patch('ansible_galaxy.fetch.local_file.LocalFileFetch._load_repository_archive',
                 return_value=mocker.Mock(name='mockRepoArchive'))
    mocker.patch('ansible_galaxy.fetch.local_file.LocalFileFetch._load_repository',
                 return_value=mocker.Mock(name='mockRepo'))
    local_fetch = local_file.LocalFileFetch(requirement_spec_)

    find_results = local_fetch.find()
    results = local_fetch.fetch(find_results=find_results)

    log.debug('results: %s', results)
    local_fetch.cleanup()

    # LocalFileFetch is acting directly on an existing file, so it's cleanup
    # should _not_ delete the file
    assert os.path.isfile(tmp_file_fo.name)

    # results = {'archive_path': '/tmp/tmpcle_fdtp.tar.gz', 'fetch_method': 'local_file',
    # 'custom': {'local_path': '/tmp/tmpcle_fdtp.tar.gz'},
    # 'content': {'galaxy_namespace': None, 'repo_name': '/tmp/tmpcle_fdtp.tar',
    # 'fetched_name': <Mock name='mockRepo.repository_spec.name' id='139946600228288'>}}
    assert results['archive_path'] == tmp_file_fo.name
    assert results['fetch_method'] == 'local_file'
    assert results['custom']['local_path'] == tmp_file_fo.name

    log.debug('should unlink %s here', tmp_file_fo.name)



import mimetypes
import io
import uuid


class MultiPartForm(object):
    """
    Accumulate the data to be used when posting a form.
    Borrowed from https://blog.thesparktree.com/the-unfortunately-long-story-dealing-with
    """

    def __init__(self):
        self.form_fields = []
        self.files = []
        self.boundary = '--------------------------%s' % uuid.uuid4().hex
        return

    def get_content_type(self):
        return 'multipart/form-data; boundary=%s' % self.boundary

    def add_field(self, name, value):
        """Add a simple field to the form data."""
        self.form_fields.append((name, value))
        return

    def add_file(self, fieldname, filename, fileHandle, mimetype=None):
        """Add a file to be uploaded."""
        body = fileHandle.read()
        if mimetype is None:
            mimetype = mimetypes.guess_type(filename)[0] or 'application/octet-stream'
        self.files.append((fieldname, filename, mimetype, body))
        return

    def get_binary(self):
        """Return a binary buffer containing the form data, including attached files."""
        part_boundary = '--' + self.boundary

        binary = io.BytesIO()
        needsCLRF = False
        # Add the form fields
        for name, value in self.form_fields:
            if needsCLRF:
                binary.write(str.encode('\r\n', 'utf-8'))
            needsCLRF = True

            block = [part_boundary,
                     'Content-Disposition: form-data; name="%s"' % name,
                     '',
                     value
                     ]
            binary.write(str.encode('\r\n'.join(block), 'utf-8'))

        # Add the files to upload
        for field_name, filename, content_type, body in self.files:
            if needsCLRF:
                binary.write(str.encode('\r\n', 'utf-8'))
            needsCLRF = True

            block = [part_boundary,
                     str('Content-Disposition: file; name="%s"; filename="%s"' %
                         (field_name, filename)),
                     'Content-Type: %s' % content_type,
                     ''
                     ]
            binary.write(str.encode('\r\n'.join(block), 'utf-8'))
            binary.write(str.encode('\r\n', 'utf-8'))
            binary.write(body)

        # add closing boundary marker,
        binary.write(str.encode('\r\n--' + self.boundary + '--\r\n', 'utf-8'))
        return binary



import logging

import pytest

from ansible_galaxy.utils.version import normalize_version_string


log = logging.getLogger(__name__)


version_string_to_strip_data = [
    'v1.2.3',
    'V1.2.3',
    'v11.1',

    # TODO: should these be stripped?
    # 'vRfd11',
    # 'v.1'
]

version_string_unmodified_data = [
    '1.2.3',
    'vV11.12',
    'veryBestVersion',
    'v',
    '1.2.4.v1',
    'vAlpha-1.1.2',
    '40.1.0.0.4v2',
    # TODO: should these be stripped?
    'vRfd11',
    'v.1'
]


@pytest.fixture(scope='module',
                params=version_string_to_strip_data)
def version_string_to_strip(request):
    yield request.param


def test_normalize_version_string_to_strip(version_string_to_strip):
    res = normalize_version_string(version_string_to_strip)
    log.debug('res: %s version_string: %s', res, version_string_to_strip)
    assert res != version_string_to_strip
    # assert not res.startswith('v')
    # assert not res.startswith('V')


@pytest.fixture(scope='module',
                params=version_string_unmodified_data)
def version_string_unmodified(request):
    yield request.param


def test_normalize_version_string_unmodified(version_string_unmodified):
    res = normalize_version_string(version_string_unmodified)
    log.debug('res: %s version_string: %s', res, version_string_unmodified)
    assert res == version_string_unmodified




import logging
import os

from ansible_galaxy.build import Build, BuildStatuses
from ansible_galaxy import collection_info

log = logging.getLogger(__name__)


def ensure_output_dir(output_path):
    if not os.path.isdir(output_path):
        log.debug('Creating output_path: %s', output_path)
        os.makedirs(output_path)
    return output_path


def _build(galaxy_context,
           build_context,
           display_callback=None):

    results = {}

    log.debug('build_context: %s', build_context)

    collection_path = build_context.collection_path
    collection_info_file_path = os.path.join(collection_path, collection_info.COLLECTION_INFO_FILENAME)

    results['collection_path'] = collection_path
    results['info_file_path'] = collection_info_file_path
    results['errors'] = []

    info = None

    try:
        with open(collection_info_file_path, 'r') as info_fd:
            info = collection_info.load(info_fd)

            log.debug('info: %s', info)
    except IOError as e:
        log.error('Error loading the %s at %s: %s', collection_info.COLLECTION_INFO_FILENAME, collection_info_file_path, e)
        results['errors'].append('Error loading the %s at %s: %s' % (collection_info.COLLECTION_INFO_FILENAME,
                                                                     collection_info_file_path, e))

    if not info:
        results['success'] = False
        results['errors'].append('There was no collection info in %s' % collection_info_file_path)
        return results

    builder = Build(build_context=build_context,
                    collection_info=info)

    ensure_output_dir(build_context.output_path)

    build_results = builder.run(display_callback=display_callback)

    log.debug('build_results: %s', build_results)

    # results here include the builder results and... ?
    results['build_results'] = build_results

    log.debug('build action results: %s', results)

    if build_results.status == BuildStatuses.success:
        results['success'] = True
        return results

    results['success'] = False
    return results


def build(galaxy_context,
          build_context,
          display_callback=None):
    '''Run _list action and return an exit code suitable for process exit'''

    results = _build(galaxy_context,
                     build_context,
                     display_callback=display_callback)

    log.debug('cli build action results: %s', results)

    if results['errors']:
        for error in results['errors']:
            display_callback(error)

    if results['success']:
        return os.EX_OK  # 0

    return os.EX_SOFTWARE  # 70



import logging

from ansible_galaxy.actions import version

log = logging.getLogger(__name__)


def test_version_data():
    config_path = '/dev/null/some/faux/mazer.yml'
    cli_version = '1.2.3'
    args = ['/bin/mazer', 'version']

    res = version.version_data(config_file_path=config_path,
                               cli_version=cli_version,
                               argv=args)

    log.debug('res: %s', res)

    assert isinstance(res, dict)

    for field in version.VERSION_FIELDS:
        assert field in res

    assert res['config_file'] == config_path
    assert res['version'] == cli_version
    assert res['executable_location'] == args[0]


def test_version():
    config_path = '/dev/null/some/faux/mazer.yml'
    cli_version = '1.2.3'
    # args = ['/bin/mazer', 'version']

    display_items = []

    def callback(*args):
        for arg in args:
            display_items.append(arg)

    res = version.version(config_file_path=config_path,
                          cli_version=cli_version,
                          display_callback=callback)

    assert res == 0

    log.debug('display_items: %s', display_items)

    lines = display_items[0].splitlines()
    assert 'mazer' in lines[0]


def test_version_repr():
    config_path = '/dev/null/some/faux/mazer.yml'
    cli_version = '1.2.3'
    args = ['/bin/mazer', 'version']

    data = version.version_data(config_file_path=config_path,
                                cli_version=cli_version,
                                argv=args)

    res = version.version_repr(data)
    assert 'mazer.yml' in res


def test_version_repr_empty_data():
    data = {}
    res = version.version_repr(data)

    assert 'name =' in res
    assert 'version =' in res




import logging
import os

from ansible_galaxy import matchers
from ansible_galaxy.models.galaxy_namespace import GalaxyNamespace

log = logging.getLogger(__name__)


def get_namespace_paths(content_path):
    # TODO: abstract this a bit?  one to make it easier to mock, but also
    #       possibly to prepare for nested dirs, multiple paths, various
    #       filters/whitelist/blacklist/excludes, caching, or respecting
    #       fs ordering, etc
    try:
        # TODO: filter on any rules for what a namespace path looks like
        #       may one being 'somenamespace.somename' (a dot sep ns and name)
        #
        namespace_paths = os.listdir(content_path)
    except OSError as e:
        log.exception(e)
        log.warning('The content path %s did not exist so no content or repositories were found.',
                    content_path)
        namespace_paths = []

    return namespace_paths


def installed_namespace_iterator(galaxy_context,
                                 match_filter=None):

    namespace_match_filter = match_filter or matchers.MatchAll()

    content_path = galaxy_context.content_path

    namespace_paths = get_namespace_paths(content_path)

    log.debug('Looking for namespaces in %s', content_path)
    for namespace_path in namespace_paths:
        namespace_full_path = os.path.join(content_path, namespace_path)

        collection_namespace = GalaxyNamespace(namespace=namespace_path,
                                               path=namespace_full_path)

        if namespace_match_filter(collection_namespace):
            log.debug('Found namespace "%s"', collection_namespace.namespace)
            yield collection_namespace


class InstalledNamespaceDatabase(object):
    def __init__(self, installed_context=None):
        self.installed_context = installed_context

    def select(self, namespace_match_filter=None):
        namespace_match_filter = namespace_match_filter or matchers.MatchAll()

        installed_namespaces = installed_namespace_iterator(self.installed_context,
                                                            match_filter=namespace_match_filter)

        for matched_namespace in installed_namespaces:
            yield matched_namespace



import datetime
import logging

import attr

from ansible_galaxy.models.install_info import InstallInfo
from ansible_galaxy.models.role_metadata import RoleMetadata

log = logging.getLogger(__name__)


@attr.s(frozen=True)
class InstallationResults(object):
    install_info_path = attr.ib()
    install_info = attr.ib(type=InstallInfo)

    installed_to_path = attr.ib()
    installed_datetime = attr.ib(type=datetime.datetime)

    installed_files = attr.ib(factory=list)

    meta_main = attr.ib(type=RoleMetadata, default=None)




import attr


@attr.s(frozen=True)
class ContentItem(object):
    '''A content item from a collection, for ex a role or plugin.

    For ex, 'testing.ansible-testing-content' will have multiple content
    items. One for each role or type of plugins.
    '''
    namespace = attr.ib()
    name = attr.ib()

    content_item_type = attr.ib(default=None)
    version = attr.ib(default=None)
    path = attr.ib(default=None)

    @property
    def label(self):
        return '%s.%s' % (self.namespace, self.name)



import json
import logging
import os

log = logging.getLogger(__name__)

# use get_spdx() to ref the spdx_license info, so it
# only loaded from disk once
_SPDX_LICENSES = None


def load_spdx():
    cwd = os.path.dirname(os.path.abspath(__file__))
    license_path = os.path.join(cwd, '..', 'data', 'spdx_licenses.json')
    try:
        with open(license_path, 'r') as fo:
            return json.load(fo)
    except EnvironmentError as exc:
        log.warning('Unable to open %s to load the list of acceptable open source licenses: %s',
                    license_path, exc)
        log.exception(exc)
        return {}


def get_spdx():
    global _SPDX_LICENSES

    if not _SPDX_LICENSES:
        _SPDX_LICENSES = load_spdx()

    return _SPDX_LICENSES



# attempt to abstract (tar) archive handling a bit

# would like a easy way to extract a subdir of a tar archive to
# a directory on local fs while using relative paths

import logging
import os

from ansible_galaxy import exceptions

log = logging.getLogger(__name__)

# pass in list of tarinfo of paths to extract
# pass in a map of tar member paths -> dest paths, built separately?
#  (based on content_type and content.CONTENT_TYPE_DIR_MAP etc)


def extract_file(tar_file, file_to_extract):
    # TODO: should just be a object? ContentArchiveMember? ContentArchiveExtractData ?
    archive_member = file_to_extract['archive_member']
    dest_dir = file_to_extract['dest_dir']
    dest_filename = file_to_extract['dest_filename']
    force_overwrite = file_to_extract['force_overwrite']

    orig_name = archive_member.name
    if not archive_member.isreg() and not archive_member.issym():
        return None

    # TODO: raise from up a level in the stack?
    dest_path = os.path.join(dest_dir, dest_filename)
    # log.debug('dest_dir: %s, dest_filename: %s, dest_path: %s orig_name: %s',
    #          dest_dir, dest_filename, dest_path, orig_name)
    if os.path.exists(dest_path):
        if not force_overwrite:
            message = "The Galaxy content %s appears to already exist." % dest_path
            raise exceptions.GalaxyClientError(message)

    try:
        tar_file.getmember(archive_member.name)
    except KeyError:
        raise exceptions.GalaxyArchiveError('The archive "%s" has no file "%s"' % (tar_file.name, archive_member.name),
                                            archive_path=tar_file.name)

    # TODO: set a default owner/group
    # MAYBE TODO: pick a 'install time' and make sure all the mtime/ctime values of extracted files
    #             match the 'install time' used in .galaxy_install_info ?

    # change the tar file member name in place to just the filename ('myfoo.py') so that extract places that file in
    # dest_dir directly instead of using adding the archive path as well
    # like '$dest_dir/archive-roles/library/myfoo.py'
    archive_member.name = dest_filename

    # log.debug('tar member: %s dest_dir: %s', archive_member, dest_dir)
    tar_file.extract(archive_member, dest_dir)

    installed_path = os.path.join(dest_dir, dest_filename)

    # reset the tar info object's name attr to the origin value in
    # case something else references this
    archive_member.name = orig_name
    return installed_path


def extract_files(tar_file, files_to_extract):
    '''Process tar_file, extracting the files from files_to_extract'''

    for file_to_extract in files_to_extract:
        res = extract_file(tar_file, file_to_extract)
        if res:
            yield res




import logging

from ansible_galaxy import installed_repository_db
from ansible_galaxy import matchers

log = logging.getLogger(__name__)


# galaxy_context.content_path is empty
def test_installed_repository_db(galaxy_context):
    icdb = installed_repository_db.InstalledRepositoryDatabase(galaxy_context)

    for x in icdb.select():
        log.debug('x: %s', x)


# galaxy_context.content_path is empty
def test_installed_repository_db_match_names(galaxy_context):
    icdb = installed_repository_db.InstalledRepositoryDatabase(galaxy_context)

    match_filter = matchers.MatchNames(['foo.bar'])
    for x in icdb.select(match_filter):
        log.debug('x: %s', x)



import logging
import os
import pprint

import yaml

from ansible_galaxy.utils import chksums
from ansible_galaxy.models.collection_artifact_file import CollectionArtifactFile
from ansible_galaxy.models.collection_artifact_file_manifest import \
    CollectionArtifactFileManifest


pf = pprint.pformat
log = logging.getLogger(__name__)


COLLECTION_FILE_MANIFEST_FILENAME = 'FILES.json'


def load(data_or_file_object):
    # FIXME: This file is json now, so could use the regular json.load()
    #        (this works as well since json is subset of yaml...)
    log.debug('loading collection artifact file manifest from %s',
              pf(data_or_file_object))

    data_dict = yaml.safe_load(data_or_file_object)

    # log.debug('data: %s', pf(data_dict))
    # log.debug('data_dict: %s', data_dict)

    instance = CollectionArtifactFileManifest(**data_dict)

    log.debug('%s instance from_kwargs', type(instance))

    log.debug('collection_file_artifact_manifest: %r', instance)

    return instance


def create_file_manifest_item(file_name, file_name_in_archive):
    if os.path.isfile(file_name):
        # At the moment, 'sha256' is the only supported chksum_type for files
        ftype = 'file'
        chksum_type = 'sha256'
        chksum = chksums.sha256sum_from_path(file_name)

    if os.path.isdir(file_name):
        chksum = None
        chksum_type = None
        ftype = 'dir'

    log.debug('ftype: %s file_name: %s fnia: %s', ftype, file_name, file_name_in_archive)

    artifact_file_item = CollectionArtifactFile(src_name=file_name,
                                                # The path where the file will live inside the archive
                                                name=file_name_in_archive,
                                                ftype=ftype,
                                                chksum_type=chksum_type,
                                                chksum_sha256=chksum)

    return artifact_file_item


def gen_file_manifest_items(file_names, collection_path):
    file_names = file_names or []

    for file_name in file_names:
        file_name_in_archive = os.path.relpath(file_name, collection_path)

        artifact_file_item = create_file_manifest_item(file_name, file_name_in_archive=file_name_in_archive)

        yield artifact_file_item



import io
import logging
import json
import ssl
import sys

import pytest
import mock

from six.moves.urllib.error import HTTPError
from six import text_type

import ansible_galaxy
from ansible_galaxy import exceptions
from ansible_galaxy.models.context import GalaxyContext
from ansible_galaxy import rest_api

log = logging.getLogger(__name__)


def test_galaxy_api_init():

    gc = GalaxyContext()
    api = rest_api.GalaxyAPI(gc)

    assert isinstance(api, rest_api.GalaxyAPI)
    assert api.galaxy == gc


class FauxUrlOpenResponse(object):
    def __init__(self, url=None, body=None, status=200, data=None, info=None, redirect_url=None):
        self.status = 200
        self.data = data
        if self.data is not None:
            self.body = json.dumps(self.data)
        else:
            self.body = body or ''
        self.url = url
        self._info = info
        self.redirect_url = redirect_url

    def read(self):
        # log.debug('read: %s %s', self.body, self)
        return self.body

    def getcode(self):
        log.debug('getcode: %s', self.status)
        return self.status

    def geturl(self):
        if self.redirect_url:
            log.debug('geturl: %s self: %s', self.redirect_url, self)
            return self.redirect_url

        log.debug('geturl: %s self: %s', self.url, self)
        return self.url

    def info(self):
        # log.debug('info: %s', self._info)
        return self._info

    def __repr__(self):
        return '%s(url="%s", status=%s, info="%s", body="%s", data="%s")' % \
            (self.__class__.__name__, self.url, self.status, self._info, self.body, self.data)


class FauxUrlResponder(object):
    def __init__(self, list_of_responses=None):
        self.responses = list_of_responses or []
        self.calls = []
        # log.debug('self.responses: %s', pprint.pformat(self.responses))

    def __call__(self, *args, **kwargs):
        next_response = self.responses.pop(0)

        # set the response url to the same as the request url
        url = args[0]

        next_response.url = url

        self.calls.append({'args': args,
                           'kwargs': kwargs,
                           'response': next_response})
        # log.debug('next_response: %s', next_response)
        return next_response

    def __repr__(self):
        return '%s(responses=%s)' % (self.__class__.__name__, self.responses)


default_server_dict = {'url': 'https://galaxy-qa.ansible.com',
                       'ignore_certs': False}


def test_galaxy_api_get_server_api_version(mocker):
    mocker.patch('ansible_galaxy.rest_api.open_url',
                 new=FauxUrlResponder(
                     [
                         FauxUrlOpenResponse(data={'current_version': 'v1'}),
                     ]
                 ))

    gc = GalaxyContext(server=default_server_dict)
    api = rest_api.GalaxyAPI(gc)
    res = api._get_server_api_version()

    log.debug('res: %s', res)

    assert isinstance(res, text_type)
    assert res == 'v1'


def test_galaxy_api_get_server_api_version_not_supported_version(mocker):
    mocker.patch('ansible_galaxy.rest_api.open_url',
                 new=FauxUrlResponder(
                     [
                         FauxUrlOpenResponse(data={'current_version': 'v11.22.13beta4.preRC-16.0.0.0.42.42.37.1final'}),
                         # doesn't matter response, just need a second call
                         FauxUrlOpenResponse(data={'results': 'stuff'},
                                             url='blippyblopfakeurl'),
                     ]
                 ))

    gc = GalaxyContext(server=default_server_dict)
    api = rest_api.GalaxyAPI(gc)

    # expected to raise a client error about the server version not being supported in client
    # TODO: should be the server deciding if the client version is sufficient
    try:
        api.lookup_repo_by_name('test-namespace', 'test-repo')
    except exceptions.GalaxyClientError as e:
        log.exception(e)
        assert 'Unsupported Galaxy server API' in '%s' % e
        return

    assert False, 'Expected a GalaxyClientError about supported server apis, but didnt happen'


def test_galaxy_api_get_server_api_version_HTTPError_500(mocker):
    error_body = u'{"detail": "Stuff broke, 500 error but server response has valid json include the detail key"}'

    mocker.patch('ansible_galaxy.rest_api.open_url',
                 side_effect=HTTPError(url='http://whatever',
                                       code=500,
                                       msg='Stuff broke.',
                                       hdrs={},
                                       fp=io.StringIO(initial_value=error_body)))

    gc = GalaxyContext(server=default_server_dict)
    api = rest_api.GalaxyAPI(gc)
    try:
        api._get_server_api_version()
    except exceptions.GalaxyClientError as e:
        log.exception(e)
        # fragile, but currently return same exception so look for the right msg in args
        assert 'Failed to get data from the API server' in '%s' % e
        return

    assert False, 'Expected a GalaxyClientError here but that did not happen'


def test_galaxy_api_get_server_api_version_HTTPError_not_json(mocker):
    mocker.patch('ansible_galaxy.rest_api.open_url',
                 new=FauxUrlResponder(
                     [
                         FauxUrlOpenResponse(status=500, body='{stuff-that-is-not-valid-json'),
                     ]
                 ))

    gc = GalaxyContext(server=default_server_dict)
    api = rest_api.GalaxyAPI(gc)
    try:
        api._get_server_api_version()
    except exceptions.GalaxyClientError as e:
        log.exception(e)
        # fragile, but currently return same exception so look for the right msg in args
        assert 'Could not process data from the API server' in '%s' % e

        return

    assert False, 'Expected a GalaxyClientError here but that did not happen'


def test_galaxy_api_get_server_api_version_no_current_version(mocker):
    mocker.patch('ansible_galaxy.rest_api.open_url',
                 new=FauxUrlResponder(
                     [
                         FauxUrlOpenResponse(status=500, data={'some_key': 'some_value',
                                                               'but_not_current_value': 2}),
                     ]
                 ))

    gc = GalaxyContext(server=default_server_dict)
    api = rest_api.GalaxyAPI(gc)
    try:
        api._get_server_api_version()
    except exceptions.GalaxyClientError as e:
        log.exception(e)
        # fragile, but currently return same exception so look for the right msg in args
        assert "missing required 'current_version'" in '%s' % e

        return

    assert False, 'Expected a GalaxyClientError here but that did not happen'


galaxy_context_params = [{'server': default_server_dict,
                          'content_dir': None}]


@pytest.fixture(scope='module',
                params=galaxy_context_params)
def galaxy_api(request):
    gc = GalaxyContext(server=request.param['server'])
    # log.debug('gc: %s', gc)

    # mock the result of _get_server_api_versions here, so that we dont get the extra
    # call when calling the tests

    patcher = mock.patch('ansible_galaxy.rest_api.open_url',
                         new=FauxUrlResponder(
                             [
                                 FauxUrlOpenResponse(data={'current_version': 'v1'}),
                                 FauxUrlOpenResponse(data={'results': [{'foo1': 11, 'foo2': 12}]},
                                                     url='blippyblopfakeurl'),
                             ]
                         ))

    patcher.start()
    api = rest_api.GalaxyAPI(gc)

    # do a call so that we run g_connect and _get_server_api_versions by side effect now
    api.lookup_repo_by_name('test-namespace', 'test-repo')

    # now unpatch
    patcher.stop()

    yield api


def test_galaxy_api_properties(galaxy_api):
    log.debug('api_server: %s', galaxy_api.api_server)
    log.debug('validate_certs: %s', galaxy_api.validate_certs)

    assert galaxy_api.api_server == 'https://galaxy-qa.ansible.com'
    assert galaxy_api.validate_certs is True


def test_galaxy_api_lookup_repo_by_name(mocker, galaxy_api):
    mocker.patch('ansible_galaxy.rest_api.open_url',
                 new=FauxUrlResponder(
                     [
                         FauxUrlOpenResponse(data={'stuff': [1, 2, 3],
                                                   'results': [{'foo1': 11, 'foo2': 12},
                                                               {'other_stuff': 'blip'}]
                                                   },
                                             url='blippyblopfakeurl'),
                     ]
                 ))

    namespace = 'alikins'
    name = 'role-awx'
    res = galaxy_api.lookup_repo_by_name(namespace, name)

    log.debug('res: %s', res)

    # FIXME: lookup_content_by_name only returns the first result
    assert isinstance(res, dict)
    assert 'foo1' in res


def test_galaxy_api_lookup_repo_by_name_empty_results(mocker, galaxy_api):
    mocker.patch('ansible_galaxy.rest_api.open_url',
                 new=FauxUrlResponder(
                     [
                         FauxUrlOpenResponse(data={'stuff': [1, 2, 3],
                                                   'results': [],
                                                   },
                                             url='blippyblopfakeurl'),
                     ]
                 ))

    namespace = 'alikins'
    name = 'role-awx'
    res = galaxy_api.lookup_repo_by_name(namespace, name)

    log.debug('res: %s', res)

    # If there are no results, we expect to get back an empty dict.
    # FIXME: should probably return the full list and let the app care what that means
    assert isinstance(res, dict)
    assert res == {}


def test_galaxy_api_lookup_repo_by_name_redirect_url(mocker, galaxy_api):
    mocker.patch('ansible_galaxy.rest_api.open_url',
                 new=FauxUrlResponder(
                     [
                         FauxUrlOpenResponse(data={'stuff': [1, 2, 3],
                                                   'results': [],
                                                   },
                                             url='blippyblopfakeurl',
                                             redirect_url='https://redirectedtothisurl/foo/blorp'),
                     ]
                 ))

    namespace = 'alikins'
    name = 'role-awx'
    res = galaxy_api.lookup_repo_by_name(namespace, name)

    log.debug('res: %s', res)

    # If there are no results, we expect to get back an empty dict.
    # FIXME: should probably return the full list and let the app care what that means
    assert isinstance(res, dict)
    assert res == {}


def test_galaxy_api_lookup_repo_by_name_500_json_not_dict(mocker, galaxy_api):
    mocker.patch('ansible_galaxy.rest_api.open_url',
                 side_effect=HTTPError(url='http://whatever',
                                       code=500,
                                       msg='Stuff broke.',
                                       hdrs={},
                                       fp=io.StringIO(initial_value=u'[]')))

    try:
        galaxy_api.lookup_repo_by_name('some-test-namespace', 'some-test-name')
    except exceptions.GalaxyClientError as e:
        log.exception(e)
        log.debug(e)
        return

    assert False, 'Excepted to get a HTTPError(code=500) here but did not.'


def test_galaxy_api_lookup_repo_by_name_500_json(mocker, galaxy_api):
    error_body_text = u'{"detail": "Stuff broke, 500 error but server response has valid json include the detail key"}'

    mocker.patch('ansible_galaxy.rest_api.open_url',
                 side_effect=HTTPError(url='http://whatever',
                                       code=500,
                                       msg='Stuff broke.',
                                       hdrs={},
                                       fp=io.StringIO(initial_value=error_body_text)))

    try:
        galaxy_api.lookup_repo_by_name('some-test-namespace', 'some-test-name')
    except exceptions.GalaxyClientError as e:
        log.exception(e)
        log.debug(e)
        return

    assert False, 'Excepted to get a GalaxyClientError here but did not.'


def test_galaxy_api_lookup_repo_by_name_SSLError(mocker, galaxy_api):
    mocker.patch('ansible_galaxy.rest_api.open_url',
                 side_effect=ssl.SSLError('ssl stuff broke... good luck and godspeed.'))

    try:
        galaxy_api.lookup_repo_by_name('some-test-namespace', 'some-test-name')
    except exceptions.GalaxyClientAPIConnectionError as e:
        log.exception(e)
        log.debug(e)

        return

    assert False, 'Excepted to get a GalaxyClientAPIConnectionError here but did not.'


def test_galaxy_api_fetch_content_related_500(mocker, galaxy_api):
    error_detail_text = u'{"detail": "Stuff broke, 500 error but server response has valid json include the detail key"}'
    mocker.patch('ansible_galaxy.rest_api.open_url',
                 side_effect=HTTPError(url='http://whatever',
                                       code=500,
                                       msg='Stuff broke.',
                                       hdrs={},
                                       fp=io.StringIO(initial_value=error_detail_text)))

    # FIXME: fetch_content_related has a catch-all exception handle that returns None for any exception (including http errors etc)
    try:
        res = galaxy_api.fetch_content_related('/api/v1/repositories/56683/content/')
        log.debug('res: %s', res)
    except exceptions.GalaxyClientError:
        return

    assert False, 'Excepted to get a GalaxyClientError here but did not.'


def test_galaxy_api_fetch_content_related(mocker, galaxy_api):
    url = '/api/v1/repositories/56683/content/'
    mocker.patch('ansible_galaxy.rest_api.open_url',
                 new=FauxUrlResponder(
                     [
                         FauxUrlOpenResponse(data={'next_link': '%s?page=2' % url,
                                                   'results': ['foo1', 'foo2']
                                                   },
                                             url='blippyblopfakeurl'),
                         FauxUrlOpenResponse(data={'results': ['foo3', 'foo4']
                                                   },
                                             url='blippyblopfakeurl'),
                     ]
                 ))

    res = galaxy_api.fetch_content_related('/api/v1/repositories/56683/content/')

    log.debug('res: %s', res)

    # TODO: use real-ish response data and verify it doesnt get munged
    assert isinstance(res, list)
    assert len(res) == 4
    assert 'foo1' in res
    assert 'foo4' in res
    assert all([isinstance(x, text_type) for x in res])


def test_galaxy_api_fetch_content_related_empty_results(mocker, galaxy_api):
    url = '/api/v1/repositories/56683/content/'
    mocker.patch('ansible_galaxy.rest_api.open_url',
                 new=FauxUrlResponder(
                     [
                         FauxUrlOpenResponse(data={'next_link': '%s?page=2' % url,
                                                   'results': []
                                                   },
                                             url='blippyblopfakeurl'),
                         FauxUrlOpenResponse(data={'results': [],
                                                   },
                                             url='blippyblopfakeurl'),
                     ]
                 ))

    res = galaxy_api.fetch_content_related('/api/v1/repositories/56683/content/')

    log.debug('res: %s', res)

    assert isinstance(res, list)
    assert len(res) == 0


def test_galaxy_api_fetch_content_related_no_results(mocker, galaxy_api):
    url = '/api/v1/repositories/56683/content/'
    mocker.patch('ansible_galaxy.rest_api.open_url',
                 new=FauxUrlResponder(
                     [
                         FauxUrlOpenResponse(data={'next_link': '%s?page=2' % url,
                                                   },
                                             url='blippyblopfakeurl'),
                         FauxUrlOpenResponse(data={},
                                             url='blippyblopfakeurl'),
                     ]
                 ))

    res = galaxy_api.fetch_content_related('/api/v1/repositories/56683/content/')

    log.debug('res: %s', res)

    assert isinstance(res, list)
    assert len(res) == 0


def test_user_agent():
    res = rest_api.user_agent()
    assert res.startswith('Mazer/%s' % ansible_galaxy.__version__)
    assert sys.platform in res
    assert 'python:' in res
    assert 'ansible_galaxy' in res




import logging

import pytest

from ansible_galaxy import exceptions
from ansible_galaxy_cli import exceptions as cli_exceptions

log = logging.getLogger(__name__)


def test_galaxy_cli_error():
    galaxy_cli_error = cli_exceptions.GalaxyCliError()

    assert isinstance(galaxy_cli_error, exceptions.GalaxyError)


# TODO: paramaterize/pytext fixture
def test_galaxy_cli_error_with_message():
    msg = "there was an error in the cli"
    galaxy_cli_error = cli_exceptions.GalaxyCliError(msg)

    assert isinstance(galaxy_cli_error, exceptions.GalaxyError)

    with pytest.raises(cli_exceptions.GalaxyCliError, match=msg) as exc_info:
        raise cli_exceptions.GalaxyCliError(msg)
    log.debug("exc_info: %s", exc_info)


def test_cli_options_error():
    cli_options_error = cli_exceptions.CliOptionsError()

    assert isinstance(cli_options_error, Exception), 'instance of CliOptionsError should be subclass of Exception'




# -*- coding: utf-8 -*-
from __future__ import print_function

import logging
import re

import attr
import semantic_version
import six

from ansible_galaxy.data import spdx_licenses

log = logging.getLogger(__name__)

TAG_REGEXP = re.compile('^[a-z0-9]+$')

# match only lowercase alphanumerics or underscore without
# leading numbers or underscores or multiple consecutive underscores
# This excludes dashes '-', punct (',' or '.' etc).
# Valid: 'my_collection', 'stuff', 'the_0cho', 'abc123'
# Invalid: '_main', 'dots.inname', 'TheBigExampleCo',
#          'kilroy-_____-was_here', '¯\_(ツ)_/¯ ', '11__'
NAME_REGEXP = re.compile(r'^(?!.*__)[a-z]+[0-9a-z_]*$')

# Valid: 'abc123', 'blip'
# Invalid: '2legit2fast2sig11', '123abc', '10_9_8'
# see https://github.com/ansible/galaxy/issues/957
MATCH_LEADING_NUMBER_REGEXP = re.compile(r'^[0-9]')


def convert_none_to_empty_dict(val):
    ''' if val is None, return an empty dict'''

    # if val is not a dict or val 'None' return val
    # and let the validators raise errors later
    if val is None:
        return {}
    return val


def convert_single_to_list(val):
    '''If a single object is provided, replace with a list containing only that object'''

    if val is None:
        return []

    if not isinstance(val, list):
        return [val]

    return val


@attr.s(frozen=True)
class CollectionInfo(object):
    namespace = attr.ib(default=None)
    name = attr.ib(default=None)
    version = attr.ib(default=None)
    # license = attr.ib(default=None)
    license = attr.ib(factory=list, converter=convert_single_to_list)
    description = attr.ib(default=None)

    repository = attr.ib(default=None)
    documentation = attr.ib(default=None)
    homepage = attr.ib(default=None)
    issues = attr.ib(default=None)

    authors = attr.ib(factory=list)
    tags = attr.ib(factory=list)

    # TODO: check these are valid paths at some point
    license_file = attr.ib(default=None,
                           validator=attr.validators.optional(attr.validators.instance_of(six.string_types)))
    readme = attr.ib(default=None,
                     validator=attr.validators.optional(attr.validators.instance_of(six.string_types)))

    # Note galaxy.yml 'dependencies' field is what mazer and ansible
    # consider 'requirements'. ie, install time requirements.
    dependencies = attr.ib(factory=dict, converter=convert_none_to_empty_dict)

    def __attrs_post_init__(self):
        # validate with have something in license or license_file
        self._check_for_license_or_license_file(self.license, self.license_file)

    @property
    def label(self):
        return '%s.%s' % (self.namespace, self.name)

    @staticmethod
    def value_error(msg):
        raise ValueError("Invalid collection metadata. %s" % msg)

    @namespace.validator
    @name.validator
    @version.validator
    def _check_required(self, attribute, value):
        if value is None:
            self.value_error("'%s' is required" % attribute.name)

    @version.validator
    def _check_version_format(self, attribute, value):
        if not semantic_version.validate(value):
            self.value_error("Expecting '%s' to be in semantic version format, "
                             "instead found '%s'." % (attribute.name, value))

    def _check_for_license_or_license_file(self, license_ids, license_file):
        if license_ids or license_file:
            return

        self.value_error("Valid values for 'license' or 'license_file' are required. "
                         "But 'license' (%s) and 'license_file' (%s) were invalid." % (license_ids, license_file))

    @license.validator
    def _check_licenses(self, attribute, value):
        '''Validate that 'licenses' value is a list of valid license identifiers'''

        # load or return already loaded data
        valid_license_ids = spdx_licenses.get_spdx()

        invalid_licenses = [license_id for license_id in value if not self._is_valid_license_id(license_id, valid_license_ids)]

        if invalid_licenses:
            self.value_error("Expecting '%s' to be a list of valid SPDX license identifiers, instead found invalid license identifiers: '%s' "
                             "in 'license' value %s. "
                             "For more info, visit https://spdx.org" % (attribute.name,
                                                                        ','.join([str(license_value) for license_value in invalid_licenses]),
                                                                        value))

    @staticmethod
    def _is_valid_license_id(license_id, valid_license_ids):
        if license_id is None:
            return False

        valid = valid_license_ids.get(license_id, None)
        if valid is None:
            return False

        # license was in list, but is deprecated
        if valid and valid.get('deprecated', None):
            print("Warning: collection metadata 'license' ID '%s' is "
                  "deprecated." % license_id)

        return True

    @authors.validator
    @tags.validator
    def _check_list_type(self, attribute, value):
        if not isinstance(value, list):
            self.value_error("Expecting '%s' to be a list" % attribute.name)

    @dependencies.validator
    def _check_dependencies_type(self, attribute, value):
        if not isinstance(value, dict) or value is None:
            self.value_error("Expecting '%s' to be a dict" % attribute.name)

    @tags.validator
    def _check_keywords(self, attribute, value):
        for k in value:
            if not re.match(TAG_REGEXP, k):
                self.value_error("Expecting %s to contain lowercase alphanumeric characters only, "
                                 "instead found '%s'." % (attribute.name, k))

    @name.validator
    @namespace.validator
    def _check_name(self, _unused, value):
        if '.' in value:
            self.value_error("Expecting 'name' and 'namespace' to not include any '.' but '%s' has a '.'" % value)
        if re.match(MATCH_LEADING_NUMBER_REGEXP, value):
            self.value_error("Expecting 'name' and 'namespace' to not start with a number but '%s' did" % value)
        # since the NAME_REGEXP catches use of hyphen '-' at all, the next check doesn't need to check for leading hyphen
        if value.startswith(('_',)):
            self.value_error("Expecting 'name' and 'namespace' to not start with '_' but '%s' did" % value)
        if not re.match(NAME_REGEXP, value):
            self.value_error("Expecting 'name' and 'namespace' to contain only lowercase alphanumeric characters or '_' only but '%s' contains others" % value)
